# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
import os
from anyjson import dumps as json_dumps, loads as json_loads
from cStringIO import StringIO
from datetime import datetime
from itertools import groupby
from operator import attrgetter

# psutil
from psutil import Process

# PyYAML
import yaml

# pytz
from pytz import UTC

# Texttable
from texttable import Texttable

# Zato
from zato.common import INFO_FORMAT, MISC, ZATO_INFO_FILE
from zato.common.util import current_host

def format_connections(conns, format):
    """ Formats a list of connections according to the output format.
    """
    groups = (groupby(conns, key=attrgetter('status')))
    out = {}

    for status, items in groups:
        items = list(items)
        items.sort(key=attrgetter('raddr'))
        out_items = out.setdefault(status, [])

        for item in items:

            laddr_str = ':'.join(str(elem) for elem in item.laddr).ljust(21).decode('utf-8')
            raddr_str = ':'.join(str(elem) for elem in item.raddr).rjust(21).decode('utf-8')

            out_item = {
                'from': '{}:{}'.format(*item.laddr),
                'to': None,
                'formatted': None,
            }

            if item.raddr:
                out_item['to'] = '{}:{}'.format(*item.raddr)
                out_item['formatted'] = '{} -> {}'.format(laddr_str, raddr_str)
            else:
                out_item['formatted'] = '{}:{}'.format(*item.laddr)

            out_items.append(out_item)

    return out

def get_info(component_path, format):
    component_details = open(os.path.join(component_path, ZATO_INFO_FILE)).read()

    out = {
        'component_details': component_details,
        'component_full_path': component_path,
        'component_host': current_host(),
        'component_running': False,
        'current_time': datetime.now().isoformat(),
        'current_time_utc': datetime.utcnow().isoformat(),
        'master_proc_connections': None,
        'master_proc_pid': None,
        'master_proc_name': None,
        'master_proc_create_time': None,
        'master_proc_create_time_utc': None,
        'master_proc_username': None,
        'master_proc_workers_no': None,
        'master_proc_workers_pids': None,
    }

    master_proc_pid = None
    try:
        master_proc_pid = int(open(os.path.join(component_path, MISC.PIDFILE)).read())
    except(IOError, ValueError):
        # Ok, no such file or it's empty
        pass

    if master_proc_pid:
        out['component_running'] = True
        master_proc = Process(master_proc_pid)
        workers_pids = sorted(elem.pid for elem in master_proc.children())

        out['master_proc_connections'] = format_connections(master_proc.connections(), format)
        out['master_proc_pid'] = master_proc.pid
        out['master_proc_create_time'] = datetime.fromtimestamp(master_proc.create_time()).isoformat()
        out['master_proc_create_time_utc'] = datetime.fromtimestamp(master_proc.create_time(), UTC).isoformat()
        out['master_proc_username'] = master_proc.username()
        out['master_proc_name'] = master_proc.name()
        out['master_proc_workers_no'] = len(workers_pids)
        out['master_proc_workers_pids'] = workers_pids

        for pid in workers_pids:
            worker = Process(pid)
            out['worker_{}_create_time'.format(pid)] = datetime.fromtimestamp(worker.create_time()).isoformat()
            out['worker_{}_create_time_utc'.format(pid)] = datetime.fromtimestamp(worker.create_time(), UTC).isoformat()
            out['worker_{}_connections'.format(pid)] = format_connections(worker.connections(), format)

    return out

def format_info(value, format, cols_width=None, dumper=None):
    if format in(INFO_FORMAT.DICT, INFO_FORMAT.JSON, INFO_FORMAT.YAML):
        value['component_details'] = json_loads(value['component_details'])

    if format == INFO_FORMAT.JSON:
        return json_dumps(value)

    elif format == INFO_FORMAT.YAML:
        buff = StringIO()
        yaml.dump_all([value], default_flow_style=False, indent=4, Dumper=dumper, stream=buff)
        value = buff.getvalue()
        buff.close()

        return value

    elif format == INFO_FORMAT.TEXT:
        cols_width = (elem.strip() for elem in cols_width.split(','))
        cols_width = [int(elem) for elem in cols_width]

        table = Texttable()
        table.set_cols_width(cols_width)

        # Use text ('t') instead of auto so that boolean values don't get converted into ints
        table.set_cols_dtype(['t', 't'])

        rows = [['Key', 'Value']]
        rows.extend(sorted(value.items()))

        table.add_rows(rows)

        return table.draw()

    else:
        return value


# -*- coding: utf-8 -*-

"""
Copyright (C) 2010 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
import copy, errno, gc, inspect, json, linecache, logging, os, random, re, signal, string, threading, traceback, sys
from ast import literal_eval
from contextlib import closing
from cStringIO import StringIO
from datetime import datetime
from glob import glob
from hashlib import sha256
from importlib import import_module
from itertools import ifilter, izip, izip_longest, tee
from operator import itemgetter
from os import getuid
from os.path import abspath, isabs, join
from pprint import pprint as _pprint, PrettyPrinter
from pwd import getpwuid
from random import getrandbits
from socket import gethostname, getfqdn
from string import Template
from tempfile import NamedTemporaryFile
from threading import current_thread
from traceback import format_exc
from urlparse import urlparse

# alembic
from alembic import op

# anyjson
from anyjson import dumps, loads

# base32_crockford
from base32_crockford import encode as b32_crockford_encode

# Bunch
from bunch import Bunch, bunchify

# ConfigObj
from configobj import ConfigObj

# dateutil
from dateutil.parser import parse

# gevent
from gevent.greenlet import Greenlet
from gevent.hub import Hub

# lxml
from lxml import etree, objectify

# OpenSSL
from OpenSSL import crypto

# Paste
from paste.util.converters import asbool

# pip
from pip.download import is_archive_file, unpack_file_url

# portalocker
import portalocker

# psutil
import psutil

# pytz
import pytz

# requests
import requests

# Spring Python
from springpython.context import ApplicationContext
from springpython.remoting.http import CAValidatingHTTPSConnection
from springpython.remoting.xmlrpc import SSLClientTransport

# SQLAlchemy
from sqlalchemy.exc import IntegrityError, ProgrammingError
import sqlalchemy as sa

# Texttable
from texttable import Texttable

# validate
from validate import is_boolean, is_integer, VdtTypeError

# Zato
from zato.common import CHANNEL, DATA_FORMAT, KVDB, MISC, SECRET_SHADOW, SIMPLE_IO, soap_body_path, soap_body_xpath, TLS, \
     TRACE1, ZatoException
from zato.common.broker_message import SERVICE
from zato.common.crypto import CryptoManager
from zato.common.odb.model import HTTPSOAP, IntervalBasedJob, Job, Service
from zato.common.odb.query import _service as _service

logger = logging.getLogger(__name__)

logging.addLevelName(TRACE1, "TRACE1")

_repr_template = Template('<$class_name at $mem_loc$attrs>')
_uncamelify_re = re.compile(r'((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))')

_epoch = datetime.utcfromtimestamp(0) # Start of UNIX epoch

random.seed()

# ################################################################################################################################

TLS_KEY_TYPE = {
    crypto.TYPE_DSA: 'DSA',
    crypto.TYPE_RSA: 'RSA'
}

# ################################################################################################################################

def absolutize_path(base, path):
    """ Turns a path into an absolute path if it's relative to the base
    location. If the path is already an absolute path, it is returned as-is.
    """
    if isabs(path):
        return path

    return abspath(join(base, path))

def current_host():
    return gethostname() + '/' + getfqdn()

def pprint(obj):
    """ Pretty-print an object into a string buffer.
    """
    # Get dicts' items.
    if hasattr(obj, "items"):
        obj = sorted(obj.items())

    buf = StringIO()
    _pprint(obj, buf)

    value = buf.getvalue()
    buf.close()

    return value

def encrypt(data, priv_key, b64=True):
    """ Encrypt data using a public key derived from the private key.
    data - data to be encrypted
    priv_key - private key to use (as a PEM string)
    b64 - should the encrypted data be BASE64-encoded before being returned, defaults to True
    """

    cm = CryptoManager(priv_key=priv_key)
    cm.load_keys()

    return cm.encrypt(data, b64)

def decrypt(data, priv_key, b64=True):
    """ Decrypts data using the given private key.
    data - data to be encrypted
    priv_key - private key to use (as a PEM string)
    b64 - should the data be BASE64-decoded before being decrypted, defaults to True
    """

    cm = CryptoManager(priv_key=priv_key)
    cm.load_keys()

    return cm.decrypt(data, b64)

def get_executable():
    """ Returns the wrapper buildout uses for executing Zato commands. This has
    all the dependencies added to PYTHONPATH.
    """
    return os.path.join(os.path.dirname(sys.executable), 'py')

def get_zato_command():
    """ Returns the full path to the 'zato' command' in a buildout environment.
    """
    return os.path.join(os.path.dirname(sys.executable), 'zato')

# Based on
# http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored
class ColorFormatter(logging.Formatter):

    # TODO: Make it all configurable

    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    COLORS = {
      'WARNING': YELLOW,
      'INFO': WHITE,
      'DEBUG': BLUE,
      'CRITICAL': YELLOW,
      'ERROR': RED,
      'TRACE1': YELLOW
    }

    def __init__(self, fmt):
        self.use_color = True
        msg = self.formatter_msg(fmt, self.use_color)
        logging.Formatter.__init__(self, msg)

    def formatter_msg(self, msg, use_color=True):
        if use_color:
            msg = msg.replace("$RESET", self.RESET_SEQ).replace("$BOLD", self.BOLD_SEQ)
        else:
            msg = msg.replace("$RESET", "").replace("$BOLD", "")
        return msg

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            fore_color = 30 + self.COLORS[levelname]
            levelname_color = self.COLOR_SEQ % fore_color + levelname + self.RESET_SEQ
            record.levelname = levelname_color

        return logging.Formatter.format(self, record)

def object_attrs(_object, ignore_double_underscore, to_avoid_list, sort):
    attrs = dir(_object)

    if ignore_double_underscore:
        attrs = ifilter(lambda elem: not elem.startswith("__"), attrs)

    _to_avoid_list = getattr(_object, to_avoid_list, None) # Don't swallow exceptions
    if _to_avoid_list is not None:
        attrs = ifilter(lambda elem: not elem in _to_avoid_list, attrs)

    if sort:
        attrs = sorted(attrs)

    return attrs

def make_repr(_object, ignore_double_underscore=True, to_avoid_list='repr_to_avoid', sort=True):
    """ Makes a nice string representation of an object, suitable for logging purposes.
    """
    attrs = object_attrs(_object, ignore_double_underscore, to_avoid_list, sort)
    buff = StringIO()

    for attr in attrs:
        attr_obj = getattr(_object, attr)
        if not callable(attr_obj):
            buff.write(' ')
            buff.write('%s:`%r`' % (attr, attr_obj))

    out = _repr_template.safe_substitute(
        class_name=_object.__class__.__name__, mem_loc=hex(id(_object)), attrs=buff.getvalue())
    buff.close()

    return out

def to_form(_object):
    """ Reads public attributes of an object and creates a dictionary out of it;
    handy for providing initial data to a Django form which isn't backed by
    a true Django model.
    """
    out = {}
    attrs = object_attrs(_object, True, "repr_to_avoid", False)
    for attr in attrs:
        out[attr] = getattr(_object, attr)

    return out

def get_lb_client(lb_host, lb_agent_port, ssl_ca_certs, ssl_key_file, ssl_cert_file, timeout):
    """ Returns an SSL XML-RPC client to the load-balancer.
    """
    from zato.agent.load_balancer.client import LoadBalancerAgentClient

    agent_uri = "https://{host}:{port}/RPC2".format(host=lb_host, port=lb_agent_port)

    # See the 'Problems with XML-RPC over SSL' thread for details
    # https://lists.springsource.com/archives/springpython-users/2011-June/000480.html
    if sys.version_info >= (2, 7):
        class Python27CompatTransport(SSLClientTransport):
            def make_connection(self, host):
                return CAValidatingHTTPSConnection(
                    host, strict=self.strict, ca_certs=self.ca_certs,
                    keyfile=self.keyfile, certfile=self.certfile, cert_reqs=self.cert_reqs,
                    ssl_version=self.ssl_version, timeout=self.timeout)
        transport = Python27CompatTransport
    else:
        transport = None

    return LoadBalancerAgentClient(
        agent_uri, ssl_ca_certs, ssl_key_file, ssl_cert_file, transport=transport, timeout=timeout)

def tech_account_password(password_clear, salt):
    return sha256(password_clear+ ':' + salt).hexdigest()

def new_cid():
    """ Returns a new 128-bit correlation identifier. It's *not* safe to use the ID
    for any cryptographical purposes, it's only meant to be used as a conveniently
    formatted ticket attached to each of the requests processed by Zato servers.
    Changed in 2.0: The number is now 28 characters long not 40, like in previous versions.
    """
    # The number below (27) needs to be kept in sync with zato.common.log_message.CID_LENGTH.
    # There is nothing special in the 'K' prefix, it's just so that a CID always
    # begins with a letter and 'K' seems like something
    # that can't be taken for some other ASCII letter (e.g. is it Z or 2 etc.)
    return 'K{0:0>27}'.format(b32_crockford_encode(getrandbits(127) + (1 << 127)))

def get_config(repo_location, config_name, bunchified=True):
    """ Returns the configuration object. Will load additional user-defined config files,
    if any are available at all.
    """
    conf = ConfigObj(os.path.join(repo_location, config_name))
    conf = bunchify(conf) if bunchified else conf
    conf.user_config_items = {}

    # user_config is new in 2.0
    user_config = conf.get('user_config')
    if user_config:
        for name, path in user_config.items():
            if not isabs(path):
                path = os.path.expanduser(path)
            if not isabs(path):
                path = os.path.normpath(os.path.join(repo_location, path))
            if not os.path.exists(path):
                logger.warn('User config not found `%s`, name:`%s`', path, name)
            else:
                user_conf = ConfigObj(path)
                user_conf = bunchify(user_conf) if bunchified else user_conf
                conf.user_config_items[name] = user_conf

    return conf

def _get_ioc_config(location, config_class):
    """ Instantiates an Inversion of Control container from the given location
    if the location exists at all.
    """
    stat = os.stat(location)
    if stat.st_size:
        config = config_class(location)
    else:
        config = None

    return config

def get_app_context(config):
    """ Returns the Zato's Inversion of Control application context.
    """
    ctx_class_path = config['spring']['context_class']
    ctx_class_path = ctx_class_path.split('.')
    mod_name, class_name = '.'.join(ctx_class_path[:-1]), ctx_class_path[-1:][0]
    mod = import_module(mod_name)
    class_ = getattr(mod, class_name)()
    return ApplicationContext(class_)

def get_crypto_manager(repo_location, app_context, config, load_keys=True):
    """ Returns a tool for crypto manipulations.
    """
    crypto_manager = app_context.get_object('crypto_manager')

    priv_key_location = config['crypto']['priv_key_location']
    cert_location = config['crypto']['cert_location']
    ca_certs_location = config['crypto']['ca_certs_location']

    priv_key_location = absolutize_path(repo_location, priv_key_location)
    cert_location = absolutize_path(repo_location, cert_location)
    ca_certs_location = absolutize_path(repo_location, ca_certs_location)

    crypto_manager.priv_key_location = priv_key_location
    crypto_manager.cert_location = cert_location
    crypto_manager.ca_certs_location = ca_certs_location

    if load_keys:
        crypto_manager.load_keys()

    return crypto_manager

def get_current_user():
    return getpwuid(getuid()).pw_name

def service_name_from_impl(impl_name):
    """ Turns a Zato internal service's implementation name into a shorter
    service name
    """
    return impl_name.replace('server.service.internal.', '')

def deployment_info(method, object_, timestamp, fs_location, remote_host='', remote_user=''):
    """ Returns a JSON document containing information who deployed a service
    onto a server, where from and when it was.
    """
    return {
        'method': method,
        'object': object_,
        'timestamp': timestamp,
        'fs_location':fs_location,
        'remote_host': remote_host,
        'remote_user': remote_user,
        'current_host': current_host(),
        'current_user': get_current_user(),
    }

def get_body_payload(body):
    body_children_count = body[0].countchildren()

    if body_children_count == 0:
        body_payload = None
    elif body_children_count == 1:
        body_payload = body[0].getchildren()[0]
    else:
        body_payload = body[0].getchildren()

    return body_payload

def payload_from_request(cid, request, data_format, transport):
    """ Converts a raw request to a payload suitable for usage with SimpleIO.
    """
    if request is not None:
        if data_format == DATA_FORMAT.XML:
            if transport == 'soap':
                if isinstance(request, objectify.ObjectifiedElement):
                    soap = request
                else:
                    soap = objectify.fromstring(request)
                body = soap_body_xpath(soap)
                if not body:
                    raise ZatoException(cid, 'Client did not send the [{}] element'.format(soap_body_path))
                payload = get_body_payload(body)
            else:
                if isinstance(request, objectify.ObjectifiedElement):
                    payload = request
                else:
                    payload = objectify.fromstring(request)
        elif data_format in(DATA_FORMAT.DICT, DATA_FORMAT.JSON):
            if not request:
                return ''
            if isinstance(request, basestring) and data_format == DATA_FORMAT.JSON:
                payload = loads(request)
            else:
                payload = request
        else:
            payload = request
    else:
        payload = request

    return payload

def is_python_file(name):
    """ Is it a Python file we can import Zato services from?
    """
    for suffix in('py', 'pyw'):
        if name.endswith(suffix):
            return True

def fs_safe_name(value):
    return re.sub('[{}]'.format(string.punctuation + string.whitespace), '_', value)

def fs_safe_now():
    """ Returns a UTC timestamp with any characters unsafe for filesystem names
    removed.
    """
    return fs_safe_name(str(datetime.utcnow()))

class _DummyLink(object):
    """ A dummy class for staying consistent with pip's API in certain places
    below.
    """
    def __init__(self, url):
        self.url = url

def decompress(archive, dir_name):
    """ Decompresses an archive into a directory, the directory must already exist.
    """
    unpack_file_url(_DummyLink('file:' + archive), dir_name)

def visit_py_source(dir_name):
    for pattern in('*.py', '*.pyw'):
        glob_path = os.path.join(dir_name, pattern)
        for py_path in sorted(glob(glob_path)):
            yield py_path

def _os_remove(path):
    """ A helper function so it's easier to mock it in unittests.
    """
    return os.remove(path)

def hot_deploy(parallel_server, file_name, path, delete_path=True, notify=True):
    """ Hot-deploys a package if it looks like a Python module or archive.
    """
    if is_python_file(file_name) or is_archive_file(file_name):

        logger.debug('About to hot-deploy [{}]'.format(path))
        now = datetime.utcnow()
        di = dumps(deployment_info('hot-deploy', file_name, now.isoformat(), path))

        # Insert the package into the DB ..
        package_id = parallel_server.odb.hot_deploy(
            now, di, file_name, open(path, 'rb').read(), parallel_server.id)

        # .. and optionally notify all the servers they're to pick up a delivery
        if notify:
            parallel_server.notify_new_package(package_id)

        if delete_path:
            _os_remove(path)

        return package_id

    else:
        logger.warn('Ignoring {}'.format(path))

# As taken from http://wiki.python.org/moin/SortingListsOfDictionaries
def multikeysort(items, columns):
    comparers = [((itemgetter(col[1:].strip()), -1) if col.startswith('-') else (itemgetter(col.strip()), 1)) for col in columns]

    def comparer(left, right):
        for fn, mult in comparers:
            result = cmp(fn(left), fn(right))
            if result:
                return mult * result
        else:
            return 0
    return sorted(items, cmp=comparer)

# From http://docs.python.org/release/2.7/library/itertools.html#recipes
def grouper(n, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return izip_longest(fillvalue=fillvalue, *args)

def translation_name(system1, key1, value1, system2, key2):
    return KVDB.SEPARATOR.join((KVDB.TRANSLATION, system1, key1, value1, system2, key2))

def dict_item_name(system, key, value):
    return KVDB.SEPARATOR.join((system, key, value))

# From http://docs.python.org/release/2.7/library/itertools.html#recipes
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return izip(a, b)

def from_local_to_utc(dt, tz_name, dayfirst=True):
    """ What is the UTC time given the local time and the timezone's name?
    """
    if not isinstance(dt, datetime):
        dt = parse(dt, dayfirst=dayfirst)

    dt = pytz.timezone(tz_name).localize(dt)
    utc_dt = pytz.utc.normalize(dt.astimezone(pytz.utc))
    return utc_dt

def from_utc_to_local(dt, tz_name):
    """ What is the local time in the user-provided time zone name?
    """
    if not isinstance(dt, datetime):
        dt = parse(dt)

    local_tz = pytz.timezone(tz_name)
    dt = local_tz.normalize(dt.astimezone(local_tz))
    return dt

# ##############################################################################

def _utcnow():
    """ See zato.common.util.utcnow for docstring.
    """
    return datetime.utcnow()

def utcnow():
    """ A thin wrapper around datetime.utcnow added so that tests can mock it
    out and return their own timestamps at will.
    """
    return _utcnow()

def _now(tz):
    """ See zato.common.util.utcnow for docstring.
    """
    return datetime.now(tz)

def now(tz=None):
    """ A thin wrapper around datetime.now added so that tests can mock it
    out and return their own timestamps at will.
    """
    return _now(tz)

def datetime_to_seconds(dt):
    """ Converts a datetime object to a number of seconds since UNIX epoch.
    """
    return (dt - _epoch).total_seconds()

# ##############################################################################

def clear_locks(kvdb, server_token, kvdb_config=None, decrypt_func=None):
    """ Clears out any KVDB locks held by Zato servers.
    """
    if kvdb_config:
        kvdb.config = kvdb_config

    if decrypt_func:
        kvdb.decrypt_func = decrypt_func

    kvdb.init()

    for name in kvdb.conn.keys('{}*{}*'.format(KVDB.LOCK_PREFIX, server_token)):
        value = kvdb.conn.get(name)
        logger.debug('Deleting lock:[{}], value:[{}]'.format(name, value))
        kvdb.conn.delete(name)

    kvdb.close()


# Inspired by http://stackoverflow.com/a/9283563
def uncamelify(s, separator='-', elem_func=unicode.lower):
    """ Converts a CamelCaseName into a more readable one, e.g.
    will turn ILikeToReadWSDLDocsNotReallyNOPENotMeQ into
    i-like-to-read-wsdl-docs-not-really-nope-not-me-q or a similar one,
    depending on the value of separator and elem_func.
    """
    return separator.join(elem_func(elem) for elem in re.sub(_uncamelify_re, r' \1', s).split())

def get_component_name(prefix='parallel'):
    """ Returns a name of the component issuing a given request so it's possible
    to trace which Zato component issued it.
    """
    return '{}/{}/{}/{}'.format(prefix, current_host(), os.getpid(), current_thread().name)

def dotted_getattr(o, path):
    return reduce(getattr, path.split('.'), o)


def get_service_by_name(session, cluster_id, name):
    logger.debug('Looking for name:[{}] in cluster_id:[{}]'.format(name, cluster_id))
    return _service(session, cluster_id).\
           filter(Service.name==name).\
           one()

def add_startup_jobs(cluster_id, odb, stats_jobs):
    """ Adds one of the interval jobs to the ODB. Note that it isn't being added
    directly to the scheduler because we want users to be able to fine-tune the job's
    settings.
    """
    with closing(odb.session()) as session:
        for item in stats_jobs:

            try:
                service_id = get_service_by_name(session, cluster_id, item['service'])[0]

                now = datetime.utcnow()

                job = Job(None, item['name'], True, 'interval_based', now, item.get('extra', '').encode('utf-8'),
                          cluster_id=cluster_id, service_id=service_id)

                kwargs = {}
                for name in('seconds', 'minutes'):
                    if name in item:
                        kwargs[name] = item[name]

                ib_job = IntervalBasedJob(None, job, **kwargs)

                session.add(job)
                session.add(ib_job)
                session.commit()
            except(IntegrityError, ProgrammingError), e:
                session.rollback()
                logger.debug('Caught an expected error, carrying on anyway, e:[%s]', format_exc(e).decode('utf-8'))

def hexlify(item):
    """ Returns a nice hex version of a string given on input.
    """
    return ' '.join([elem1+elem2 for (elem1, elem2) in grouper(2, item.encode('hex'))])

def validate_input_dict(cid, *validation_info):
    """ Checks that input belongs is one of allowed values.
    """
    for key_name, key, source in validation_info:
        if not source.has(key):
            msg = 'Invalid {}:[{}]'.format(key_name, key)
            log_msg = '{} (attrs: {})'.format(msg, source.attrs)

            logger.warn(log_msg)
            raise ZatoException(cid, msg)

# ################################################################################################################################

# Code below taken from tripod https://github.com/shayne/tripod/blob/master/tripod/sampler.py and slightly modified
# under the terms of LGPL (see LICENSE.txt file for details).

class SafePrettyPrinter(PrettyPrinter, object):
    def format(self, obj, context, maxlevels, level):
        try:
            return super(SafePrettyPrinter, self).format(
                obj, context, maxlevels, level)
        except Exception:
            return object.__repr__(obj)[:-1] + ' (bad repr)>', True, False


def spformat(obj, depth=None):
    return SafePrettyPrinter(indent=1, width=76, depth=depth).pformat(obj)


def formatvalue(v):
    s = spformat(v, depth=1).replace('\n', '')
    if len(s) > 12500:
        s = object.__repr__(v)[:-1] + ' (really long repr)>'
    return '=' + s


def get_stack(f, with_locals=False):
    limit = getattr(sys, 'tracebacklimit', None)

    frames = []
    n = 0
    while f is not None and (limit is None or n < limit):
        lineno, co = f.f_lineno, f.f_code
        name, filename = co.co_name, co.co_filename
        args = inspect.getargvalues(f)

        linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        if line:
            line = line.strip()
        else:
            line = None

        frames.append((filename, lineno, name, line, f.f_locals, args))
        f = f.f_back
        n += 1
    frames.reverse()

    out = []
    for filename, lineno, name, line, localvars, args in frames:
        out.append(' File "%s", line %d, in %s' % (filename, lineno, name))
        if line:
            out.append(' %s' % line.strip())

        if with_locals:
            args = inspect.formatargvalues(formatvalue=formatvalue, *args)
            out.append('\n Arguments: %s%s' % (name, args))

        if with_locals and localvars:
            out.append(' Local variables:\n')
            try:
                reprs = spformat(localvars)
            except Exception:
                reprs = "failed to format local variables"
            out += [' ' + l for l in reprs.splitlines()]
            out.append('')
    return '\n'.join(out)

# ################################################################################################################################

def get_threads_traceback(pid):
    result = {}
    id_name = dict([(th.ident, th.name) for th in threading.enumerate()])

    for thread_id, frame in sys._current_frames().items():
        key = '{}:{}'.format(pid, id_name.get(thread_id, '(No name)'))
        result[key] = get_stack(frame, True)

    return result

def get_greenlets_traceback(pid):
    result = {}
    for item in gc.get_objects():
        if not isinstance(item, (Greenlet, Hub)):
            continue
        if not item:
            continue

        key = '{}:{}'.format(pid, repr(item))
        result[key] = ''.join(get_stack(item.gr_frame, True))

    return result

def dump_stacks(*ignored):
    pid = os.getpid()

    table = Texttable()
    table.set_cols_width((30, 90))
    table.set_cols_dtype(['t', 't'])

    rows = [['Proc:Thread/Greenlet', 'Traceback']]

    rows.extend(sorted(get_threads_traceback(pid).items()))
    rows.extend(sorted(get_greenlets_traceback(pid).items()))

    table.add_rows(rows)
    logger.info('\n' + table.draw())

# Taken from https://stackoverflow.com/a/16589622
def get_full_stack():
    exc = sys.exc_info()[0]
    stack = traceback.extract_stack()[:-1]  # last one would be full_stack()
    if not exc is None:  # i.e. if an exception is present
        del stack[-1]       # remove call of full_stack, the printed exception
                            # will contain the caught exception caller instead
    trc = 'Traceback (most recent call last):\n'
    stackstr = trc + ''.join(traceback.format_list(stack))

    if not exc is None:
        stackstr += '  ' + traceback.format_exc().lstrip(trc)

    return stackstr

def register_diag_handlers():
    """ Registers diagnostic handlers dumping stacks, threads and greenlets on receiving a signal.
    """
    signal.signal(signal.SIGURG, dump_stacks)

# ################################################################################################################################

def parse_extra_into_dict(lines, convert_bool=True):
    """ Creates a dictionary out of key=value lines.
    """
    _extra = {}

    if lines:
        extra = ';'.join(lines.splitlines())

        for line in extra.split(';'):
            original_line = line
            if line:
                line = line.split('=')
                if not len(line) == 2:
                    raise ValueError('Each line must be a single key=value entry, not [{}]'.format(original_line))

                key, value = line
                value = value.strip()

                if convert_bool:
                    try:
                        value = is_boolean(value)
                    except VdtTypeError:
                        # It's cool, not a boolean
                        pass

                try:
                    value = is_integer(value)
                except VdtTypeError:
                    # OK, not an integer
                    pass

                # Could be a dict or another simple type then
                try:
                    value = literal_eval(value)
                except Exception:
                    pass

                # OK, let's just treat it as string
                _extra[key.strip()] = value

    return _extra

# ################################################################################################################################

# Taken from http://plumberjack.blogspot.cz/2009/09/how-to-treat-logger-like-output-stream.html

class LoggerWriter:
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level

    def write(self, message):
        if message != '\n':
            self.logger.log(self.level, message)

# ################################################################################################################################

def validate_xpath(expr):
    """ Evaluates an XPath expression thus confirming it is correct.
    """
    etree.XPath(expr)
    return True

# ################################################################################################################################

# Taken from http://grodola.blogspot.com/2014/04/reimplementing-netstat-in-cpython.html
def is_port_taken(port):
    for conn in psutil.net_connections(kind='tcp'):
        if conn.laddr[1] == port and conn.status == psutil.CONN_LISTEN:
            return True
    return False

# ################################################################################################################################

def get_haproxy_pidfile(component_dir):
    json_config = json.loads(open(os.path.join(component_dir, 'config', 'repo', 'lb-agent.conf')).read())
    return os.path.abspath(os.path.join(component_dir, json_config['pid_file']))

def store_pidfile(component_dir):
    open(os.path.join(component_dir, MISC.PIDFILE), 'w').write('{}'.format(os.getpid()))

# ################################################################################################################################

def get_kvdb_config_for_log(config):
    config = copy.deepcopy(config)
    if config.shadow_password_in_logs:
        config.password = SECRET_SHADOW
    return config

def has_redis_sentinels(config):
    return asbool(config.get('use_redis_sentinels', False))

# ################################################################################################################################

def alter_column_nullable_false(table_name, column_name, default_value, column_type):
    column = sa.sql.table(table_name, sa.sql.column(column_name))
    op.execute(column.update().values({column_name:default_value}))
    op.alter_column(table_name, column_name, type_=column_type, existing_type=column_type, nullable=False)

# ################################################################################################################################

def validate_tls_from_payload(payload, is_key=False):
    with NamedTemporaryFile(prefix='zato-tls-') as tf:
        tf.write(payload)
        tf.flush()

        pem = open(tf.name).read()

        cert_info = crypto.load_certificate(crypto.FILETYPE_PEM, pem)
        cert_info = sorted(dict(cert_info.get_subject().get_components()).items())
        cert_info = '; '.join('{}={}'.format(k, v) for k, v in cert_info)

        if is_key:
            key_info = crypto.load_privatekey(crypto.FILETYPE_PEM, pem)
            key_info = '{}; {} bits'.format(TLS_KEY_TYPE[key_info.type()], key_info.bits())
            return '{}; {}'.format(key_info, cert_info)
        else:
            return cert_info

get_tls_from_payload = validate_tls_from_payload

def get_tls_full_path(root_dir, component, info):
    return os.path.join(root_dir, component, fs_safe_name(info) + '.pem')

def get_tls_ca_cert_full_path(root_dir, info):
    return get_tls_full_path(root_dir, TLS.DIR_CA_CERTS, info)

def get_tls_key_cert_full_path(root_dir, info):
    return get_tls_full_path(root_dir, TLS.DIR_KEYS_CERTS, info)

def store_tls(root_dir, payload, is_key=False):

    # Raises exception if it's not really a certificate.
    info = get_tls_from_payload(payload, is_key)

    pem_file_path = get_tls_full_path(root_dir, TLS.DIR_KEYS_CERTS if is_key else TLS.DIR_CA_CERTS, info)
    pem_file = open(pem_file_path, 'w')

    try:
        portalocker.lock(pem_file, portalocker.LOCK_EX)

        pem_file.write(payload)
        pem_file.close()

        os.chmod(pem_file_path, 0o640)

        return pem_file_path

    except portalocker.LockException:
        pass # It's OK, something else is doing the same thing right now

# ################################################################################################################################

def replace_private_key(orig_payload):
    if isinstance(orig_payload, basestring):
        for item in TLS.BEGIN_END:
            begin = '-----BEGIN {}PRIVATE KEY-----'.format(item)
            if begin in orig_payload:
                end = '-----END {}PRIVATE KEY-----'.format(item)

                begin_last_idx = orig_payload.find(begin) + len(begin) + 1
                end_preceeding_idx = orig_payload.find(end) -1

                return orig_payload[0:begin_last_idx] + SECRET_SHADOW + orig_payload[end_preceeding_idx:]

    # No private key at all in payload
    return orig_payload

# ################################################################################################################################

def delete_tls_material_from_fs(server, info, full_path_func):
    try:
        os.remove(full_path_func(server.tls_dir, info))
    except OSError, e:
        if e.errno == errno.ENOENT:
            # It's ok - some other worker must have deleted it already
            pass
        else:
            raise

# ################################################################################################################################

def ping_solr(config):
    result = urlparse(config.address)
    requests.get('{}://{}{}'.format(result.scheme, result.netloc, config.ping_path))

# ################################################################################################################################

def ping_odoo(conn):
    user_model = conn.get_model('res.users')
    ids = user_model.search([('login', '=', conn.login)])
    user_model.read(ids[0], ['login'])['login']

# ################################################################################################################################

class StaticConfig(Bunch):
    def __init__(self, path):
        super(StaticConfig, self).__init__()
        self.read(path)

    def read(self, path):
        for item in os.listdir(path):
            f = open(os.path.join(path, item))
            value = f.read()
            f.close()
            self[item] = value

# ################################################################################################################################

def add_scheduler_jobs(server, spawn=True):
    for(id, name, is_active, job_type, start_date, extra, service_name, _,
        _, weeks, days, hours, minutes, seconds, repeats, cron_definition)\
            in server.odb.get_job_list(server.cluster_id):

        if is_active:
            job_data = Bunch({'id':id, 'name':name, 'is_active':is_active,
                'job_type':job_type, 'start_date':start_date,
                'extra':extra, 'service':service_name, 'weeks':weeks,
                'days':days, 'hours':hours, 'minutes':minutes,
                'seconds':seconds, 'repeats':repeats,
                'cron_definition':cron_definition})
            server.singleton_server.scheduler.create_edit('create', job_data, spawn=spawn)

# ################################################################################################################################

def get_basic_auth_credentials(auth):

    if not auth:
        return None, None

    prefix = 'Basic '
    if not auth.startswith(prefix):
        return None, None

    _, auth = auth.split(prefix)
    auth = auth.strip().decode('base64')

    return auth.split(':', 1)

# ################################################################################################################################

def parse_tls_channel_security_definition(value):
    if not value:
        raise ValueError('No definition given `{}`'.format(repr(value)))

    for line in value.splitlines():
        line = line.strip()

        if not line:
            continue

        if not '=' in line:
            raise ValueError("Line `{}` has no '=' key/value separator".format(line))

        # It's possible we will have multiple '=' symbols.
        sep_index = line.find('=')
        key, value = line[:sep_index], line[sep_index+1:]

        if not key:
            raise ValueError('Key missing in line `{}`'.format(line))

        if not value:
            raise ValueError('Value missing in line `{}`'.format(line))

        yield 'HTTP_X_ZATO_TLS_{}'.format(key.upper()), value

# ################################################################################################################################

def get_http_json_channel(name, service, cluster, security):
    return HTTPSOAP(None, '{}.json'.format(name), True, True, 'channel', 'plain_http', None, '/zato/json/{}'.format(name),
        None, '', None, SIMPLE_IO.FORMAT.JSON, service=service, cluster=cluster, security=security)

def get_http_soap_channel(name, service, cluster, security):
    return HTTPSOAP(None, name, True, True, 'channel', 'soap', None, '/zato/soap', None, name, '1.1',
        SIMPLE_IO.FORMAT.XML, service=service, cluster=cluster, security=security)

# ################################################################################################################################

def startup_service_payload_from_path(name, value, repo_location):
    """ Reads payload from a local file. Abstracted out to ease in testing.
    """
    orig_path = value.replace('file://', '')
    if not os.path.isabs(orig_path):
        path = os.path.normpath(os.path.join(repo_location, orig_path))
    else:
        path = orig_path

    try:
        payload = open(path).read()
    except Exception, e:
        logger.warn(
            'Could not open payload path:`%s` `%s`, skipping startup service:`%s`, e:`%s`', orig_path, path, name, format_exc(e))
    else:
        return payload

def invoke_startup_services(
        source, key, fs_server_config, repo_location, broker_client=None, service_name=None, skip_include=True, worker_store=None):
    """ Invoked when we are the first worker and we know we have a broker client and all the other config ready
    so we can publish the request to execute startup services. In the worst case the requests will get back to us but it's
    also possible that other workers are already running. In short, there is no guarantee that any server or worker in particular
    will receive the requests, only that there will be exactly one.
    """
    for name, payload in fs_server_config.get(key, {}).items():

        if service_name:

            # We are to skip this service:
            if skip_include:
                if name == service_name:
                    continue

            # We are to include this service only, any other is rejected
            else:
                if name != service_name:
                    continue

        if payload.startswith('file://'):
            payload = startup_service_payload_from_path(name, payload, repo_location)
            if not payload:
                continue

        cid = new_cid()

        msg = {}
        msg['action'] = SERVICE.PUBLISH.value
        msg['service'] = name
        msg['payload'] = payload
        msg['cid'] = cid
        msg['channel'] = CHANNEL.STARTUP_SERVICE

        if broker_client:
            broker_client.invoke_async(msg)
        else:
            worker_store.on_message_invoke_service(msg, msg['channel'], msg['action'])



# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# Core publish/subscribe features

# stdlib
from datetime import datetime, timedelta
from json import dumps, loads
from logging import getLogger
from sys import maxint
from traceback import format_exc
import logging

# gevent
from gevent.lock import RLock

# Zato
from zato.common import PUB_SUB, ZATO_NONE, ZATO_NOT_GIVEN
from zato.common.kvdb import LuaContainer
from zato.common.pubsub import lua
from zato.common.util import datetime_to_seconds, make_repr, new_cid

# ################################################################################################################################

logger = getLogger(__name__)

# ################################################################################################################################

class HasAutoRepr(object):
    def __repr__(self):
        return make_repr(self)

# ################################################################################################################################

class PubSubException(Exception):
    """ A base class for pub/sub exception.
    """

class PermissionDenied(PubSubException):
    """ Raised when an attempt is made to make use of a topic a client is not allowed to access.
    """

class ItemFull(Exception):
    """ Raised when either a topic or a consumer's queue is full.
    """
    def __init__(self, msg, item, max_depth):
        self.msg = msg
        self.item = item
        self.max_depth = max_depth

# ################################################################################################################################

class Topic(HasAutoRepr):
    def __init__(self, name, is_active=True, is_fifo=PUB_SUB.DEFAULT_IS_FIFO, max_depth=PUB_SUB.DEFAULT_MAX_DEPTH):
        self.name = name
        self.is_active = is_active
        self.is_fifo = is_fifo
        self.max_depth = max_depth

# ################################################################################################################################

class Message(object):
    """ A published or received message.
    """
    def __init__(self, payload='', topic=None, mime_type=PUB_SUB.DEFAULT_MIME_TYPE, priority=PUB_SUB.DEFAULT_PRIORITY,
             expiration=PUB_SUB.DEFAULT_EXPIRATION, msg_id=None, producer=None, creation_time_utc=None,
             expire_at_utc=None):
        self.payload = payload
        self.topic = topic
        self.mime_type = mime_type
        self.priority = priority # In 1-9 range where 9 is top priority
        self.msg_id = msg_id or new_cid()
        self.producer = producer
        self.expiration = expiration
        self.creation_time_utc = creation_time_utc or datetime.utcnow()
        self.expire_at_utc = expire_at_utc or (self.creation_time_utc + timedelta(seconds=self.expiration))

        # These two, in local timezone, are used by web-admin.
        self.creation_time = None
        self.expire_at = None

        self.id = None # Used by frontend only
        self.payload_html = None # Used by frontend only

    def __repr__(self):
        return make_repr(self)

    def to_dict(self):
        if isinstance(self.creation_time_utc, basestring):
            creation_time_utc = self.creation_time_utc
        else:
            creation_time_utc = self.creation_time_utc.isoformat()

        if isinstance(self.expire_at_utc, basestring):
            expire_at_utc = self.expire_at_utc
        else:
            expire_at_utc = self.expire_at_utc.isoformat()

        return {
            'topic': self.topic,
            'mime_type': self.mime_type,
            'priority': self.priority,
            'msg_id': self.msg_id,
            'creation_time_utc': creation_time_utc,
            'expire_at_utc': expire_at_utc,
            'expiration': self.expiration,
            'producer': self.producer
        }

    def to_json(self):
        return dumps(self.to_dict())

# ################################################################################################################################

class PubCtx(HasAutoRepr):
    """ A set of data describing what to publish.
    """
    def __init__(self, client_id=None, topic=None, msg=None):
        self.client_id = client_id
        self.topic = topic
        self.msg = msg

# ################################################################################################################################

class SubCtx(HasAutoRepr):
    """ Subscription context - what to subscribe to.
    """
    def __init__(self, client_id=None, topics=None):
        self.client_id = client_id
        self.topics = topics or []

# ################################################################################################################################

class GetCtx(HasAutoRepr):
    """ A set of data describing where to fetch messages from.
    """
    def __init__(self, sub_key=None, max_batch_size=PUB_SUB.DEFAULT_GET_MAX_BATCH_SIZE, is_fifo=PUB_SUB.DEFAULT_IS_FIFO,
                   get_format=PUB_SUB.GET_FORMAT.OBJECT.id):
        self.sub_key = sub_key
        self.max_batch_size = max_batch_size
        self.is_fifo = is_fifo # Fetch in FIFO or LIFO order
        self.get_format = get_format

# ################################################################################################################################

class AckCtx(HasAutoRepr):
    """ A set of data describing an acknowledge_delete of a message fetched.
    """
    def __init__(self, sub_key=None, msg_ids=None):
        self.sub_key = sub_key
        self.msg_ids = msg_ids or []

    def append(self, msg_id):
        self.msg_ids.append(msg_id)

    def __repr__(self):
        return make_repr(self)

# ################################################################################################################################

class RejectCtx(HasAutoRepr):
    """ A set of data describing a rejection of one or more messages.
    """
    def __init__(self, sub_key=None, msg_ids=None):
        self.sub_key = sub_key
        self.msg_ids = msg_ids or []

    def append(self, msg_id):
        self.msg_ids.append(msg_id)

# ################################################################################################################################

class Client(HasAutoRepr):
    """ Either a subscriber or publisher.
    """
    def __init__(self, id, name, is_active=True):
        self.id = id
        self.name = name
        self.is_active = is_active

class Consumer(Client):
    """ Pub/sub consumer.
    """
    def __init__(self, id, name, is_active=True, sub_key=None, max_depth=PUB_SUB.DEFAULT_MAX_BACKLOG,
            delivery_mode=PUB_SUB.DELIVERY_MODE.PULL.id, callback_id='', callback_name=None, callback_type=ZATO_NOT_GIVEN):
        super(Consumer, self).__init__(id, name, is_active)
        self.sub_key = sub_key
        self.max_depth = max_depth
        self.delivery_mode = delivery_mode
        self.callback_id = callback_id
        self.callback_name = callback_name
        self.callback_type = callback_type

# ################################################################################################################################

class PubSub(object):
    """ An entry point to the pub/sub mechanism. Must be partly subclassed by concrete implementation classes.
    """
    def __init__(self, *ignored_args, **ignored_kwargs):
        self.logger = getLogger('zato_pubsub')

        # Held when modifying sets of currently known consumers and producers
        self.update_lock = RLock()

        # Held when in-flight messages are manipulated
        self.in_flight_lock = RLock()

        # All existing topics, key = topic name, value = topic object
        self.topics = {}

        # All existing consumers, key = client_id, value = client object
        self.consumers = {}

        # All existing producers, key = client_id, value = client object
        self.producers = {}

        self.sub_to_cons = {}   # String to string, key = sub_id, value = client_id using it
        self.cons_to_sub = {}   # String to string, key = client_id, value = sub_id it has

        # Consumers and topics
        self.cons_to_topic = {} # String to set, key = client_id, value = topics it's subscribed to
        self.topic_to_cons = {} # String to set, key = topic, value = client_ids subscribed to it

        # Producers and topics
        self.prod_to_topic = {} # String to set, key = client_id, value = topics it can publish to
        self.topic_to_prod = {} # String to set, key = topic, value = client_ids allowed to publish to it

        # Used by internal services
        self.default_consumer = Client(None, None)
        self.default_producer = Client(None, None)

# ################################################################################################################################

    def add_topic(self, topic):
        """ Adds a topic, with no clients attached.
        """
        with self.update_lock:
            self.topics[topic.name] = topic
            self.logger.info('Added topic `%s`', topic)

    update_topic = add_topic

# ################################################################################################################################

    def delete_topic(self, topic):
        """ Deletes a topic and any consumers or producers assigned to it.
        """
        with self.update_lock:
            deleted = self.topics[topic.name]

            consumers = self.topic_to_cons.get(topic.name, [])
            producers = self.topic_to_prod.get(topic.name, [])

            for consumer_id in list(consumers):
                self.delete_consumer(self.consumers[consumer_id], topic)

            for producer_id in list(producers):
                self.delete_producer(self.producers[producer_id], topic)

            # Implemented by subclasses
            self.delete_topic_metadata(topic)

            self.logger.info('Deleted topic `%s`', deleted)

# ################################################################################################################################

    def delete_topic_metadata(self, topic):
        """ Deletes topic's metadata. Needs to be subclasses by concrete implementations.
        """
        raise NotImplementedError('Must be implemented by subclasses')

# ################################################################################################################################

    def add_subscription(self, sub_key, client_id, topic):
        """ Adds subscription for a given sub_id, client and topic. Must be called with self.update_lock held.
        """
        # Mapping between sub key and client and the other way around.
        self.sub_to_cons[sub_key] = client_id
        self.cons_to_sub[client_id] = sub_key

        # This consumer's topics
        topics = self.cons_to_topic.setdefault(client_id, set())
        topics.add(topic)

        # This topic's consumers
        clients = self.topic_to_cons.setdefault(topic, set())
        clients.add(client_id)

        self.logger.info('Added subscription: `%s`, `%s`, `%s`', sub_key, client_id, topic)

# ################################################################################################################################

    def add_producer(self, client, topic):
        """ Adds information that this client can publish to the topic. 
        """
        with self.update_lock:

            # Topics for this producer
            topics = self.prod_to_topic.setdefault(client.id, set())
            topics.add(topic.name)

            # Producers for this topic
            producers = self.topic_to_prod.setdefault(topic.name, set())
            producers.add(client.id)

            self.producers[client.id] = client

            self.logger.info('Added producer `%s` for topic:`%s`', client, topic)

    def update_producer(self, client, topic):
        """ Updates a producer.
        """
        # Currently we can only switch is_active flag on/off
        with self.update_lock:
            self.producers[client.id].is_active = client.is_active
            self.logger.info('Updated producer `%s` for topic:`%s`', client, topic)

    def delete_producer(self, client, topic):
        """ Deletes an association between a producer and a topic.
        """
        with self.update_lock:

            # Topics for this producer
            topics = self.prod_to_topic.pop(client.id, set())
            topics.remove(topic.name)

            # Producers for this topic
            producers = self.topic_to_prod.get(topic.name, set())
            producers.remove(client.id)

            # That was the last producer for this topic so let's delete the topic from this dict
            # so other parts of the code can assume that if a topic doesn't exist in it, it means there is no producer
            # for the given topic name.
            if not self.topic_to_prod[topic.name]:
                del self.topic_to_prod[topic.name]

            del self.producers[client.id]
            self.delete_producer_metadata(client)

            self.logger.info('Deleted producer `%s` for topic:`%s`', client, topic)

    def delete_producer_metadata(self, client):
        """ Deletes producer's metadata. Needs to be subclasses by concrete implementations.
        """
        raise NotImplementedError('Must be implemented by subclasses')

# ################################################################################################################################

    def add_consumer(self, client, topic):
        """ Adds information that this client can publish to the topic. 
        """
        with self.update_lock:
            topics = self.cons_to_topic.setdefault(client.id, set())
            topics.add(topic.name)

            consumers = self.topic_to_cons.setdefault(topic.name, set())
            consumers.add(client.id)

            self.consumers[client.id] = client

            self.sub_to_cons[client.sub_key] = client.id
            self.cons_to_sub[client.id] = client.sub_key

            self.logger.info('Added consumer `%s` for topic:`%s`', client, topic)

    def update_consumer(self, client, topic):
        """ Updates a consumer.
        """
        with self.update_lock:
            self.consumers[client.id].is_active = client.is_active
            self.consumers[client.id].max_depth = client.max_depth
            self.consumers[client.id].delivery_mode = client.delivery_mode
            self.consumers[client.id].callback_id = client.callback_id

            self.logger.info('Updated consumer `%s` for topic:`%s`', client, topic)

    def delete_consumer(self, client, topic):
        """ Deletes an association between a consumer and a topic.
        """
        with self.update_lock:
            topics = self.cons_to_topic.pop(client.id, set())
            topics.remove(topic.name)

            consumers = self.topic_to_cons.get(topic.name, set())
            consumers.remove(client.id)

            # That was the last consumer for this topic so let's delete the topic from this dict
            # so other parts of the code can assume that if a topic doesn't exist in it, it means there is no consumer
            # for the given topic name.
            if not self.topic_to_cons[topic.name]:
                del self.topic_to_cons[topic.name]

            del self.consumers[client.id]
            del self.sub_to_cons[client.sub_key]
            del self.cons_to_sub[client.id]

            self.delete_consumer_metadata(client)

            self.logger.info('Deleted consumer `%s` for topic:`%s`', client, topic)

    def delete_consumer_metadata(self, client):
        """ Deletes consumer's metadata. Needs to be subclasses by concrete implementations.
        """
        raise NotImplementedError('Must be implemented by subclasses')

# ################################################################################################################################

    def _not_implemented(self, *ignored_args, **ignored_kwargs):
        raise NotImplementedError('Must be overridden in subclasses')

    publish = subscribe = get = acknowledge_delete = reject = create = _not_implemented

# ################################################################################################################################

class RedisPubSub(PubSub, LuaContainer):
    """ Publish/subscribe based on Redis.
    """
    # Main public API
    LUA_PUBLISH = 'lua-publish'
    LUA_GET_FROM_CONSUMER_QUEUE = 'lua-get-from-consumer-queue'
    LUA_REJECT = 'lua-reject'
    LUA_ACK_DELETE = 'lua-ack-delete'

    # Background tasks
    LUA_DELETE_EXPIRED_TOPIC = 'lua-delete-expired-topic'
    LUA_DELETE_EXPIRED_CONSUMER = 'lua-delete-expired-consumer'
    LUA_MOVE_TO_TARGET_QUEUES = 'lua-move-to-target-queues'

    # Message browsing
    LUA_GET_MESSAGE_LIST = 'lua-get-message-list'

    # Message deleting
    LUA_DELETE_FROM_TOPIC = 'lua-delete-from-topic'

# ################################################################################################################################

    def __init__(self, kvdb, key_prefix='zato:pubsub:'):
        super(RedisPubSub, self).__init__()
        self.kvdb = kvdb
        self.lua_programs = {}

        self.MSG_IDS_PREFIX = '{}{}'.format(key_prefix, 'zset:msg-ids:{}')
        self.BACKLOG_FULL_KEY = '{}{}'.format(key_prefix, 'hash:backlog-full')
        self.CONSUMER_MSG_IDS_PREFIX = '{}{}'.format(key_prefix, 'list:consumer:msg-ids:{}')
        self.CONSUMER_IN_FLIGHT_IDS_PREFIX = '{}{}'.format(key_prefix, 'set:consumer:in-flight:ids:{}')
        self.CONSUMER_IN_FLIGHT_DATA_PREFIX = '{}{}'.format(key_prefix, 'hash:consumer:in-flight:data:{}')
        self.MSG_VALUES_KEY = '{}{}'.format(key_prefix, 'hash:msg-values')
        self.MSG_METADATA_KEY = '{}{}'.format(key_prefix, 'hash:msg-metadata')
        self.MSG_EXPIRE_AT_KEY = '{}{}'.format(key_prefix, 'hash:msg-expire-at') # In UTC
        self.UNACK_COUNTER_KEY = '{}{}'.format(key_prefix, 'hash:unack-counter')
        self.LAST_PUB_TIME_KEY = '{}{}'.format(key_prefix, 'hash:last-pub-time') # In UTC
        self.LAST_SEEN_CONSUMER_KEY = '{}{}'.format(key_prefix, 'hash:last-seen-consumer') # In UTC
        self.LAST_SEEN_PRODUCER_KEY = '{}{}'.format(key_prefix, 'hash:last-seen-producer') # In UTC

        self.add_lua_program(self.LUA_PUBLISH, lua.lua_publish)
        self.add_lua_program(self.LUA_MOVE_TO_TARGET_QUEUES, lua.lua_move_to_target_queues)
        self.add_lua_program(self.LUA_GET_FROM_CONSUMER_QUEUE, lua.lua_get_from_cons_queue)
        self.add_lua_program(self.LUA_REJECT, lua.lua_reject)
        self.add_lua_program(self.LUA_ACK_DELETE, lua.lua_ack_delete)
        self.add_lua_program(self.LUA_DELETE_EXPIRED_TOPIC, lua.lua_delete_expired_topic)
        self.add_lua_program(self.LUA_DELETE_EXPIRED_CONSUMER, lua.lua_delete_expired_consumer)
        self.add_lua_program(self.LUA_GET_MESSAGE_LIST, lua.lua_get_message_list)
        self.add_lua_program(self.LUA_DELETE_FROM_TOPIC, lua.lua_delete_from_topic)

# ################################################################################################################################

    def ping(self):
        """ Pings the pub/sub backend.
        """
        return self.kvdb.ping()

# ################################################################################################################################

    def validate_sub_key(self, sub_key):
        """ Returns a client_id by its matching subscription key or raises PubSubException if sub_key could not be found.
        Must be called with self.update_lock held.
        """
        with self.update_lock:
            # Grab the client's ID if it's a valid subscription key.
            if not self.sub_to_cons.get(sub_key):
                msg = 'Invalid sub_key `{}`'.format(sub_key)
                self.logger.warn(msg)
                raise PubSubException(msg)

            return True

# ################################################################################################################################

    # Overridden from the base class

    def delete_topic_metadata(self, topic):
        self.kvdb.hdel(self.LAST_PUB_TIME_KEY, topic.name)

    def delete_consumer_metadata(self, client):
        self.kvdb.hdel(self.LAST_SEEN_CONSUMER_KEY, client.id)

    def delete_producer_metadata(self, client):
        self.kvdb.hdel(self.LAST_SEEN_PRODUCER_KEY, client.id)

# ################################################################################################################################

    def _raise_cant_publish_error(self, ctx):
        raise PermissionDenied("Permision denied. Can't publish to `{}`".format(ctx.topic))

    def publish(self, ctx):
        """ Publishes a message on a selected topic.
        """
        # Note that the client always receives the same response but logs contain details
        with self.update_lock:
            if not ctx.topic in self.topics:
                self.logger.warn('Permision denied. No such topic `%s`, publisher `%s`', ctx.topic, ctx.client_id)
                self._raise_cant_publish_error(ctx)

            if not ctx.client_id in self.topic_to_prod.get(ctx.topic, []):
                self.logger.warn('Permision denied. Producer `%s` cannot publish to `%s`', ctx.client_id, ctx.topic)
                self._raise_cant_publish_error(ctx)

            if not self.topics[ctx.topic].is_active:
                self.logger.warn('Topic `%s` is not active. Producer `%s`.', ctx.topic, ctx.client_id)
                self._raise_cant_publish_error(ctx)

            if not self.producers[ctx.client_id].is_active:
                self.logger.warn('Producer `%s` is not active. Producer `%s`.', ctx.client_id, ctx.topic)
                self._raise_cant_publish_error(ctx)

        if self.get_topic_depth(ctx.topic) >= self.topics[ctx.topic].max_depth:
            self.logger.warn('Topic full, `%s`, max depth `%s`', ctx.topic, self.topics[ctx.topic].max_depth)
            raise ItemFull('Topic full', ctx.topic, self.topics[ctx.topic].max_depth)

        id_key = self.MSG_IDS_PREFIX.format(ctx.topic)

        # Each message will carry information what topic it's intended for
        ctx.msg.topic = ctx.topic

        # The score is built by prefixing the number of milliseconds since UNIX epoch with message's priority. Hence higher
        # priority messages will get higher score whereas messages of equal priority will be still scored according
        # to their time of being published. But in the latter case this is still approximate with a high rate of publications
        # so if guarantees regarding the order of messages are required the messages should be arranged
        # in sequences on client side.

        now_seconds = datetime_to_seconds(datetime.utcnow())
        score = '{}{}'.format(ctx.msg.priority, now_seconds)

        try:
            self.run_lua(
                self.LUA_PUBLISH, [
                    id_key, self.MSG_VALUES_KEY, self.MSG_METADATA_KEY, self.MSG_EXPIRE_AT_KEY, self.LAST_PUB_TIME_KEY,
                      self.LAST_SEEN_PRODUCER_KEY],
                    [score, ctx.msg.msg_id, ctx.msg.expire_at_utc.isoformat(), ctx.msg.payload, ctx.msg.to_json(),
                       ctx.topic, datetime.utcnow().isoformat(), ctx.client_id])
        except Exception, e:
            self.logger.error('Pub error `%s`', format_exc(e))
            raise
        else:
            self.logger.info('Published `%s` to `%s`, exp `%s`', ctx.msg.msg_id, ctx.topic, ctx.msg.expire_at_utc.isoformat())
            return ctx

# ################################################################################################################################

    def subscribe(self, ctx, sub_key=None):
        """ Subscribes the client to one or more topics, or topic patterns. Returns subscription key
        to use in subsequent calls to fetch messages by.
        """
        sub_key = sub_key or new_cid()
        with self.update_lock:
            for topic in ctx.topics:
                # TODO: Resolve topic here - it can be a pattern instead of a concrete name
                self.add_subscription(sub_key, ctx.client_id, topic)

        self.logger.info('Client `%s` sub to topics `%s`', ctx.client_id, ', '.join(ctx.topics))
        return sub_key

# ################################################################################################################################

    def get(self, ctx):
        self.logger.debug('Get by sub_key `%s`', ctx.sub_key)

        with self.in_flight_lock:
            with self.update_lock:

                # Ignoring the result, we just check if this sub_key is valid
                self.validate_sub_key(ctx.sub_key)

                # Don't let the client get new messages if the depth of the in-flight queue reached its maximum.
                consumer = self.get_consumer_by_sub_key(ctx.sub_key)
                if self.get_consumer_queue_in_flight_depth(ctx.sub_key) >= consumer.max_depth:
                    self.logger.warn('In-flight queue full, `%s`, max depth `%s`', consumer.name, consumer.max_depth)
                    raise ItemFull('In-flight queue full ({max_depth}/{max_depth})'.format(max_depth=consumer.max_depth),
                        consumer.name, consumer.max_depth)

                # Now that the client is known to be a valid one we can get all their messages
                cons_queue = self.CONSUMER_MSG_IDS_PREFIX.format(ctx.sub_key)
                cons_in_flight_ids = self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(ctx.sub_key)
                cons_in_flight_data = self.CONSUMER_IN_FLIGHT_DATA_PREFIX.format(ctx.sub_key)

                messages = self.run_lua(
                    self.LUA_GET_FROM_CONSUMER_QUEUE,
                    [cons_queue, cons_in_flight_ids, cons_in_flight_data, self.LAST_SEEN_CONSUMER_KEY,
                         self.MSG_METADATA_KEY, self.MSG_VALUES_KEY],
                    [ctx.max_batch_size, datetime.utcnow().isoformat(), self.sub_to_cons[ctx.sub_key]])

                self.logger.debug('Get messages `%s`:`%r`', ctx.sub_key, messages)

                for msg in messages:

                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug('Get result: sub_key `%s`, msg `%s`', ctx.sub_key, msg)
                    else:
                        self.logger.info('Get result: sub_key `%s`, metadata `%s`', ctx.sub_key, msg[1])

                    payload = msg[0][0] if msg[0] else None
                    metadata = loads(msg[1][0])

                    if ctx.get_format == PUB_SUB.GET_FORMAT.JSON.id:
                        yield {'payload': payload, 'metadata':metadata}
                    else:
                        yield Message(payload=payload, **metadata)

# ################################################################################################################################

    def acknowledge_delete(self, ctx, is_delete=False):
        """ Consumer confirms and accepts one or more message.
        """
        # Check comment in self.reject on why validating sub_key alone suffices.
        self.validate_sub_key(ctx.sub_key)

        cons_in_flight_ids = self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(ctx.sub_key)
        cons_in_flight_data = self.CONSUMER_IN_FLIGHT_DATA_PREFIX.format(ctx.sub_key)
        cons_queue = self.CONSUMER_MSG_IDS_PREFIX.format(ctx.sub_key)

        result = self.run_lua(
            self.LUA_ACK_DELETE,
            keys=[
                cons_in_flight_ids, cons_in_flight_data, self.UNACK_COUNTER_KEY, self.MSG_VALUES_KEY, 
                self.MSG_EXPIRE_AT_KEY, self.MSG_METADATA_KEY, cons_queue],
            args=[int(is_delete)] + ctx.msg_ids) 

        self.logger.info(
            '%s: result `%s` for sub_key `%s` with msgs `%s`',
              'Del from queue' if is_delete else 'Ack', result, ctx.sub_key, ', '.join(ctx.msg_ids))

        return result

    def reject(self, ctx):
        """ Rejects a set of messages for a given consumer. The messages will be placed back onto consumer's queue
        and delivered again at a later time.
        """
        # We only validate that this subscription key is valid. Each consumer has its own
        # hashmap of on-flight messages so if they know the sub key, meaning they know the username/password,
        # if any, the worst they can do is to attempt to reject IDs that don't exist.
        # But as long as they don't know each other's subscription keys they can't reject each other's messages.
        self.validate_sub_key(ctx.sub_key)

        cons_queue = self.CONSUMER_MSG_IDS_PREFIX.format(ctx.sub_key)
        cons_in_flight_ids = self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(ctx.sub_key)
        cons_in_flight_data = self.CONSUMER_IN_FLIGHT_DATA_PREFIX.format(ctx.sub_key)

        result = self.run_lua(self.LUA_REJECT, keys=[cons_queue, cons_in_flight_ids, cons_in_flight_data], args=ctx.msg_ids) 

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.info(
                'Reject result `%s` for `%s` with `%s`', result, ctx.sub_key, ', '.join(ctx.msg_ids))

        return result

# ################################################################################################################################

    def delete_expired(self):
        """ Deletes expired messages. For each topic and its subscribers a Lua program is called to find expired
        messages and delete all traces of them.
        """
        with self.update_lock:

            now = datetime.utcnow().isoformat()

            # Delete from topic queues - needed if there are no subscribers
            for topic in self.topics:
                keys = [self.MSG_IDS_PREFIX.format(topic), self.MSG_VALUES_KEY, self.MSG_METADATA_KEY, self.MSG_EXPIRE_AT_KEY]

                self.logger.info('Delete expired (topic) `%r` for keys `%s`',
                    self.run_lua(self.LUA_DELETE_EXPIRED_TOPIC, keys=keys, args=[now]), keys)

            # Delete from consumer queues
            for consumer in self.cons_to_topic:
                sub_key = self.cons_to_sub[consumer]
                consumer_msg_ids = self.CONSUMER_MSG_IDS_PREFIX.format(sub_key)
                consumer_in_flight_ids = self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(sub_key)

                keys = [consumer_msg_ids, consumer_in_flight_ids, self.MSG_VALUES_KEY, self.MSG_EXPIRE_AT_KEY, 
                        self.UNACK_COUNTER_KEY]

                self.logger.info('Delete expired (consumer) `%r` for keys `%s`',
                    self.run_lua(self.LUA_DELETE_EXPIRED_CONSUMER, keys=keys, args=[now]), keys)

# ############################################################################################################################

    def move_to_target_queues(self):
        """ Invoked periodically in order to fetch data sent to a topic and move it to each consumer's queue.
        """
        # TODO: We currently deliver messages to each consumer. However, we also need to support 
        # the delivery to only one consumer chosen randomly from each of the subscribed ones.

        with self.update_lock:

            out = []

            for topic in self.topic_to_prod:

                source_queue = self.MSG_IDS_PREFIX.format(topic)
                topic_info = self.topics[topic]

                # There are as many keys as there are arguments - items on idx 3 and above are related and come in pairs.

                # Keys the Lua program will operate on
                keys = []
                keys.append(source_queue)
                keys.append(self.BACKLOG_FULL_KEY)
                keys.append(self.UNACK_COUNTER_KEY)

                # Lua program's args
                args = []
                args.append(int(topic_info.is_fifo)) # So it's easy to cast it to bool in Lua
                args.append(topic_info.max_depth)
                args.append(maxint)

                consumers = self.topic_to_cons.get(topic, [])
                if consumers:
                    for consumer in consumers:
                        sub_key = self.cons_to_sub[consumer]
                        self.logger.debug('Move: Found sub `%s` for topic `%s` by consumer `%s`', sub_key, topic, consumer)

                        keys.append(self.CONSUMER_MSG_IDS_PREFIX.format(sub_key))
                        args.append(self.consumers[consumer].max_depth)

                    move_result = self.run_lua(self.LUA_MOVE_TO_TARGET_QUEUES, keys, args)
                    if move_result:
                        self.logger.info('Move: result `%s`, keys `%s`', move_result, ', '.join(keys))
                        out.append(move_result)

                else:
                    self.logger.info('Move: no consumers for topic `%s`', topic)

            return out

# ################################################################################################################################

    def get_topic_depth(self, topic):
        """ Returns current depth of a topic. Doesn't held onto any locks so by the time the data is returned to the caller
        the depth may have already changed.
        """
        return self.kvdb.zcard(self.MSG_IDS_PREFIX.format(topic))

    def get_consumer_queue_current_depth(self, sub_key):
        """ Returns current depth of a consumer's queue. Doesn't held onto any locks so by the time the data
        is returned to the caller the depth may have already changed.
        """
        return self.kvdb.llen(self.CONSUMER_MSG_IDS_PREFIX.format(sub_key))

    def get_consumer_queue_in_flight_depth(self, sub_key):
        """ Returns current depth of an in-flight consumer's queue. Doesn't held onto any locks so by the time the data
        is returned to the caller the depth may have already changed.
        """
        return self.kvdb.scard(self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(sub_key))

    def get_consumer_by_sub_key(self, sub_key):
        return self.consumers.get(self.sub_to_cons.get(sub_key, ZATO_NONE))

    def get_consumer_by_client_id(self, client_id):
        return self.consumers.get(client_id, ZATO_NONE)

    def get_consumers_count(self, topic):
        """ Returns the number of consumers allowed to get messages from a given topic.
        """
        return len(self.topic_to_cons.get(topic, []))

    def get_last_pub_time(self, topic):
        """ Returns timestamp of the last publication to a topic, if any.
        """
        return self.kvdb.hget(self.LAST_PUB_TIME_KEY, topic)

    def get_producer_last_seen(self, client_id):
        """ Returns timestamp of the last time a producer published a message, regardless of its topic.
        """
        return self.kvdb.hget(self.LAST_SEEN_PRODUCER_KEY, client_id)

    def get_consumer_last_seen(self, client_id):
        """ Returns timestamp of the last time a consumer got a message, regardless of its topic.
        """
        return self.kvdb.hget(self.LAST_SEEN_CONSUMER_KEY, client_id)

    def get_producers_count(self, topic):
        """ Returns the number of producers allowed to publish messages to a topic.
        """
        return len(self.topic_to_prod.get(topic, []))

    def get_producer_by_sub_key(self, sub_key):
        return self.producers.get(self.sub_to_cons.get(sub_key, ZATO_NONE))

# ################################################################################################################################

    def get_consumer_queue_message_list(self, sub_key):
        """ Returns all messages from a given consumer queue by its subscriber's key.
        """
        for item in self.run_lua(
            self.LUA_GET_MESSAGE_LIST, [
                self.CONSUMER_MSG_IDS_PREFIX.format(sub_key), self.MSG_METADATA_KEY], [PUB_SUB.MESSAGE_SOURCE.CONSUMER_QUEUE.id]):
            yield Message(**loads(item))

    def get_consumer_in_flight_message_list(self, sub_key):
        """ Returns all in-flight message IDs from a given consumer queue by its subscriber's key.
        """
        return self.kvdb.smembers(self.CONSUMER_IN_FLIGHT_IDS_PREFIX.format(sub_key))

    def get_topic_message_list(self, source_name):
        """ Returns all messages from a given topic.
        """
        for item in self.run_lua(
            self.LUA_GET_MESSAGE_LIST, [
                self.MSG_IDS_PREFIX.format(source_name), self.MSG_METADATA_KEY], [PUB_SUB.MESSAGE_SOURCE.TOPIC.id]):
            yield Message(**loads(item))

    def delete_from_topic(self, source_name, msg_id):
        """ The message is deleted from a topic and if there are no subscriptions to a topic
        it's also removed from Redis keys holding the message's metadada and payload.
        """
        with self.update_lock:
            # Let's find subscriptions to this topic. If there aren't any subscriptions, we call a Lua function
            # that deletes not only the message from topic but also any other piece of information regarding it.
            has_consumers = True if source_name in self.topic_to_cons else False
            result = self.run_lua(
                self.LUA_DELETE_FROM_TOPIC,
                  [self.MSG_IDS_PREFIX.format(source_name), self.MSG_VALUES_KEY, self.MSG_METADATA_KEY, self.MSG_EXPIRE_AT_KEY],
                  [int(has_consumers), msg_id])

            self.logger.info(
                'Del from topic: result `%s`, source_name `%s`, msg_id `%s`, has_consumers `%s`',
                  result, source_name, msg_id, has_consumers)

    def delete_from_consumer_queue(self, sub_key, msg_ids):
        """ Deleting a message from a consumer's queue works exactly like acknowledging it.
        """
        return self.acknowledge_delete(AckCtx(sub_key, msg_ids), True)

    def get_message(self, msg_id):
        """ Returns payload of a message along with its metadata.
        """
        out = {'payload': self.kvdb.hget(self.MSG_VALUES_KEY, msg_id)}
        out.update(loads(self.kvdb.hget(self.MSG_METADATA_KEY, msg_id)))

        return out

    def get_callback_consumers(self):
        """ Returns these consumers who specified their messages should be delivered through callback URLs.
        """
        with self.update_lock:
            for consumer in self.consumers.values():
                if not consumer.is_active:
                    continue

                if consumer.delivery_mode == PUB_SUB.DELIVERY_MODE.CALLBACK_URL.id:
                    yield consumer

# ################################################################################################################################

class PubSubAPI(object):
    """ Provides an API for pub/sub users to publish or subscribe to messages through.
    """
    def __init__(self, impl):
        self.impl = impl
        """:type: zato.common.pubsub.RedisPubSub"""

    def publish(self, payload, topic, mime_type=None, priority=None, expiration=None, msg_id=None, expire_at=None, client_id=None):
        """ Publishes a message by a given to a given topic using a set of parameters provided.
        """
        client_id = client_id or self.get_default_producer().id
        ctx = PubCtx()
        ctx.client_id = client_id
        ctx.topic = topic
        ctx.msg = Message(
            payload, topic, mime_type or PUB_SUB.DEFAULT_MIME_TYPE, priority or PUB_SUB.DEFAULT_PRIORITY,
            expiration or PUB_SUB.DEFAULT_EXPIRATION, msg_id, self.impl.producers[client_id].name)

        return self.impl.publish(ctx)

    def subscribe(self, client_id, topics, sub_key=None):
        """ Subscribes a client to one or more topic. Returns a subscription key assigned.
        """
        if isinstance(topics, basestring):
            topics = [topics]

        # Sanity check - at this point 'topics' must be something we can iterate over
        # instead of, say, an integer.
        iter(topics)

        ctx = SubCtx()
        ctx.client_id = client_id
        ctx.topics = topics

        return self.impl.subscribe(ctx, sub_key)

    def get(self, sub_key, max_batch_size=PUB_SUB.DEFAULT_GET_MAX_BATCH_SIZE, is_fifo=PUB_SUB.DEFAULT_IS_FIFO,
            get_format=PUB_SUB.GET_FORMAT.DEFAULT.id):
        """ Gets one or more message, if any are available, for the given subscription key.
        """
        return self.impl.get(GetCtx(sub_key, max_batch_size, is_fifo, get_format))

    def acknowledge(self, sub_key, msg_ids):
        """ Acknowledges one or more message IDs for a given subscription key.
        """
        ctx = AckCtx()
        ctx.sub_key = sub_key
        ctx.msg_ids = [msg_ids] if not isinstance(msg_ids, (list, tuple, set, dict)) else msg_ids

        return self.impl.acknowledge_delete(ctx)

    def reject(self, sub_key, msg_ids):
        """ Rejects one or more message IDs for a given subscription key.
        """
        ctx = RejectCtx()
        ctx.sub_key = sub_key
        ctx.msg_ids = [msg_ids] if not isinstance(msg_ids, (list, tuple, set, dict)) else msg_ids

        return self.impl.reject(ctx)

# ################################################################################################################################

    # Admin calls below

    def add_topic(self, topic):
        return self.impl.add_topic(topic)

    def update_topic(self, topic):
        return self.impl.update_topic(topic)

    def delete_topic(self, topic):
        return self.impl.delete_topic(topic)

    def delete_consumer(self, consumer, topic):
        return self.impl.delete_consumer(consumer, topic)

    def add_producer(self, producer, topic):
        return self.impl.add_producer(producer, topic)

    def update_producer(self, producer, topic):
        return self.impl.update_producer(producer, topic)

    def delete_producer(self, producer, topic):
        return self.impl.delete_producer(producer, topic)

    def add_consumer(self, consumer, topic):
        return self.impl.add_consumer(consumer, topic)

    def update_consumer(self, consumer, topic):
        return self.impl.update_consumer(consumer, topic)

# ############################################################################################################################


    def get_topic_depth(self, topic):
        return self.impl.get_topic_depth(topic)

    def get_consumer_queue_current_depth(self, sub_key):
        return self.impl.get_consumer_queue_current_depth(sub_key)

    def get_consumer_queue_in_flight_depth(self, sub_key):
        return self.impl.get_consumer_queue_in_flight_depth(sub_key)

    def get_consumer_by_sub_key(self, sub_key):
        return self.impl.get_consumer_by_sub_key(sub_key)

    def get_consumer_by_client_id(self, client_id):
        return self.impl.get_consumer_by_client_id(client_id)

    def get_consumers_count(self, topic):
        return self.impl.get_consumers_count(topic)

    def get_producers_count(self, topic):
        return self.impl.get_producers_count(topic)

    def get_last_pub_time(self, topic):
        return self.impl.get_last_pub_time(topic)

    def get_producer_last_seen(self, client_id):
        return self.impl.get_producer_last_seen(client_id)

    def get_consumer_last_seen(self, client_id):
        return self.impl.get_consumer_last_seen(client_id)

    def get_producer_by_sub_key(self, sub_key):
        return self.producers.get(self.sub_to_cons.get(sub_key, ZATO_NONE))

# ################################################################################################################################

    def get_default_consumer(self):
        return self.impl.default_consumer

    def get_default_producer(self):
        return self.impl.default_producer

    def set_default_consumer(self, client):
        self.impl.default_consumer = client

    def set_default_producer(self, client):
        self.impl.default_producer = client

# ################################################################################################################################

    def get_consumer_queue_message_list(self, sub_key):
        return self.impl.get_consumer_queue_message_list(sub_key)

    def get_consumer_in_flight_message_list(self, sub_key):
        return self.impl.get_consumer_in_flight_message_list(sub_key)

    def get_topic_message_list(self, source_name):
        return self.impl.get_topic_message_list(source_name)

# ################################################################################################################################

    def delete_from_topic(self, topic_name, msg_id, *ignored):
        return self.impl.delete_from_topic(topic_name, msg_id)

    def delete_from_consumer_queue(self, sub_key, msg_ids):
        if isinstance(msg_ids, basestring):
            msg_ids = [msg_ids]
        return self.impl.delete_from_consumer_queue(sub_key, msg_ids if isinstance(msg_ids, list) else list(msg_ids))

# ################################################################################################################################

    def get_message(self, msg_id):
        return self.impl.get_message(msg_id)

# ################################################################################################################################

    def can_access_topic(self, client_id, topic, is_consumer):
        return topic in self.impl.cons_to_topic.get(client_id, []) if is_consumer else self.impl.prod_to_topic.get(client_id, [])

# ################################################################################################################################



# -*- coding: utf-8 -*-

"""
Copyright (C) 2011 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
from inspect import isclass

# candv
from candv import Constants as _Constants, ValueConstant as _ValueConstant

class Constants(_Constants):
    values = _Constants.constants

class ValueConstant(_ValueConstant):

    def __cmp__(self, other):
        return cmp(self.value, (other.value if isinstance(other, ValueConstant) else other))

class MESSAGE:
    MESSAGE_TYPE_LENGTH = 4
    TOKEN_LENGTH = 32
    TOKEN_START = MESSAGE_TYPE_LENGTH
    TOKEN_END = MESSAGE_TYPE_LENGTH + TOKEN_LENGTH
    PAYLOAD_START = MESSAGE_TYPE_LENGTH + TOKEN_LENGTH
    NULL_TOKEN = '0' * TOKEN_LENGTH

class MESSAGE_TYPE:
    TO_SINGLETON = b'0000'
    TO_PARALLEL_ANY = b'0001'
    TO_PARALLEL_ALL = b'0002'

    TO_AMQP_PUBLISHING_CONNECTOR_ALL = b'0003'
    TO_AMQP_CONSUMING_CONNECTOR_ALL = b'0004'
    TO_AMQP_CONNECTOR_ALL = b'0005'
    
    TO_JMS_WMQ_PUBLISHING_CONNECTOR_ALL = b'0006'
    TO_JMS_WMQ_CONSUMING_CONNECTOR_ALL = b'0007'
    TO_JMS_WMQ_CONNECTOR_ALL = b'0008'
    
    TO_ZMQ_PUBLISHING_CONNECTOR_ALL = b'0009'
    TO_ZMQ_CONSUMING_CONNECTOR_ALL = b'0010'
    TO_ZMQ_CONNECTOR_ALL = b'0011'

    USER_DEFINED_START = b'5000'

TOPICS = {
    MESSAGE_TYPE.TO_SINGLETON: b'/zato/to-singleton',

    MESSAGE_TYPE.TO_PARALLEL_ANY: b'/zato/to-parallel/any',
    MESSAGE_TYPE.TO_PARALLEL_ALL: b'/zato/to-parallel/all',

    MESSAGE_TYPE.TO_AMQP_PUBLISHING_CONNECTOR_ALL: b'/zato/connector/amqp/publishing/all',
    MESSAGE_TYPE.TO_AMQP_CONSUMING_CONNECTOR_ALL: b'/zato/connector/amqp/consuming/all',
    MESSAGE_TYPE.TO_AMQP_CONNECTOR_ALL: b'/zato/connector/amqp/all',

    MESSAGE_TYPE.TO_JMS_WMQ_PUBLISHING_CONNECTOR_ALL: b'/zato/connector/jms-wmq/publishing/all',
    MESSAGE_TYPE.TO_JMS_WMQ_CONSUMING_CONNECTOR_ALL: b'/zato/connector/jms-wmq/consuming/all',
    MESSAGE_TYPE.TO_JMS_WMQ_CONNECTOR_ALL: b'/zato/connector/jms-wmq/all',

    MESSAGE_TYPE.TO_ZMQ_PUBLISHING_CONNECTOR_ALL: b'/zato/connector/zmq/publishing/all',
    MESSAGE_TYPE.TO_ZMQ_CONSUMING_CONNECTOR_ALL: b'/zato/connector/zmq/consuming/all',
    MESSAGE_TYPE.TO_ZMQ_CONNECTOR_ALL: b'/zato/connector/zmq/all',
}

KEYS = {k:v.replace('/zato','').replace('/',':') for k,v in TOPICS.items()}

class SCHEDULER(Constants):
    code_start = 100000

    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')
    EXECUTE = ValueConstant('')
    JOB_EXECUTED = ValueConstant('')

class ZMQ_SOCKET(Constants):
    code_start = 100200
    CLOSE = ValueConstant('')

class SECURITY(Constants):
    code_start = 100400

    BASIC_AUTH_CREATE = ValueConstant('')
    BASIC_AUTH_EDIT = ValueConstant('')
    BASIC_AUTH_DELETE = ValueConstant('')
    BASIC_AUTH_CHANGE_PASSWORD = ValueConstant('')

    TECH_ACC_CREATE = ValueConstant('')
    TECH_ACC_EDIT = ValueConstant('')
    TECH_ACC_DELETE = ValueConstant('')
    TECH_ACC_CHANGE_PASSWORD = ValueConstant('')

    WSS_CREATE = ValueConstant('')
    WSS_EDIT = ValueConstant('')
    WSS_DELETE = ValueConstant('')
    WSS_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    OAUTH_CREATE = ValueConstant('')
    OAUTH_EDIT = ValueConstant('')
    OAUTH_DELETE = ValueConstant('')
    OAUTH_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    NTLM_CREATE = ValueConstant('')
    NTLM_EDIT = ValueConstant('')
    NTLM_DELETE = ValueConstant('')
    NTLM_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    AWS_CREATE = ValueConstant('')
    AWS_EDIT = ValueConstant('')
    AWS_DELETE = ValueConstant('')
    AWS_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    OPENSTACK_CREATE = ValueConstant('')
    OPENSTACK_EDIT = ValueConstant('')
    OPENSTACK_DELETE = ValueConstant('')
    OPENSTACK_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    APIKEY_CREATE = ValueConstant('')
    APIKEY_EDIT = ValueConstant('')
    APIKEY_DELETE = ValueConstant('')
    APIKEY_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    XPATH_SEC_CREATE = ValueConstant('')
    XPATH_SEC_EDIT = ValueConstant('')
    XPATH_SEC_DELETE = ValueConstant('')
    XPATH_SEC_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    TLS_CA_CERT_CREATE = ValueConstant('')
    TLS_CA_CERT_EDIT = ValueConstant('')
    TLS_CA_CERT_DELETE = ValueConstant('')

    # New in 2.0
    TLS_CHANNEL_SEC_CREATE = ValueConstant('')
    TLS_CHANNEL_SEC_EDIT = ValueConstant('')
    TLS_CHANNEL_SEC_DELETE = ValueConstant('')

    # New in 2.0
    TLS_KEY_CERT_CREATE = ValueConstant('')
    TLS_KEY_CERT_EDIT = ValueConstant('')
    TLS_KEY_CERT_DELETE = ValueConstant('')

class DEFINITION(Constants):
    code_start = 100600

    AMQP_CREATE = ValueConstant('')
    AMQP_EDIT = ValueConstant('')
    AMQP_DELETE = ValueConstant('')
    AMQP_CHANGE_PASSWORD = ValueConstant('')

    JMS_WMQ_CREATE = ValueConstant('')
    JMS_WMQ_EDIT = ValueConstant('')
    JMS_WMQ_DELETE = ValueConstant('')

    ZMQ_CREATE = ValueConstant('')
    ZMQ_EDIT = ValueConstant('')
    ZMQ_DELETE = ValueConstant('')

    # New in 2.0
    CASSANDRA_CREATE = ValueConstant('')
    CASSANDRA_EDIT = ValueConstant('')
    CASSANDRA_DELETE = ValueConstant('')
    CASSANDRA_CHANGE_PASSWORD = ValueConstant('')

class OUTGOING(Constants):
    code_start = 100800

    AMQP_CREATE = ValueConstant('')
    AMQP_EDIT = ValueConstant('')
    AMQP_DELETE = ValueConstant('')
    AMQP_PUBLISH = ValueConstant('')

    JMS_WMQ_CREATE = ValueConstant('')
    JMS_WMQ_EDIT = ValueConstant('')
    JMS_WMQ_DELETE = ValueConstant('')
    JMS_WMQ_SEND = ValueConstant('')

    ZMQ_CREATE = ValueConstant('')
    ZMQ_EDIT = ValueConstant('')
    ZMQ_DELETE = ValueConstant('')
    ZMQ_SEND = ValueConstant('')

    SQL_CREATE_EDIT = ValueConstant('') # Same for creating and updating the pools
    SQL_CHANGE_PASSWORD = ValueConstant('')
    SQL_DELETE = ValueConstant('')

    HTTP_SOAP_CREATE_EDIT = ValueConstant('') # Same for creating and updating
    HTTP_SOAP_DELETE = ValueConstant('')

    FTP_CREATE_EDIT = ValueConstant('') # Same for creating and updating
    FTP_DELETE = ValueConstant('')
    FTP_CHANGE_PASSWORD = ValueConstant('')

    ODOO_CREATE = ValueConstant('')
    ODOO_EDIT = ValueConstant('')
    ODOO_DELETE = ValueConstant('')
    ODOO_CHANGE_PASSWORD = ValueConstant('')

class CHANNEL(Constants):
    code_start = 101000

    AMQP_CREATE = ValueConstant('')
    AMQP_EDIT = ValueConstant('')
    AMQP_DELETE = ValueConstant('')
    AMQP_MESSAGE_RECEIVED = ValueConstant('')

    JMS_WMQ_CREATE = ValueConstant('')
    JMS_WMQ_EDIT = ValueConstant('')
    JMS_WMQ_DELETE = ValueConstant('')
    JMS_WMQ_MESSAGE_RECEIVED = ValueConstant('')

    ZMQ_CREATE = ValueConstant('')
    ZMQ_EDIT = ValueConstant('')
    ZMQ_DELETE = ValueConstant('')
    ZMQ_MESSAGE_RECEIVED = ValueConstant('')

    HTTP_SOAP_CREATE_EDIT = ValueConstant('') # Same for creating and updating
    HTTP_SOAP_DELETE = ValueConstant('')
    HTTP_SOAP_AUDIT_RESPONSE = ValueConstant('') # New in 2.0
    HTTP_SOAP_AUDIT_PATTERNS = ValueConstant('') # New in 2.0
    HTTP_SOAP_AUDIT_STATE = ValueConstant('') # New in 2.0
    HTTP_SOAP_AUDIT_CONFIG = ValueConstant('') # New in 2.0

class AMQP_CONNECTOR(Constants):
    code_start = 101200
    CLOSE = ValueConstant('')

class JMS_WMQ_CONNECTOR(Constants):
    code_start = 101400
    CLOSE = ValueConstant('')

class ZMQ_CONNECTOR(Constants):
    code_start = 101600
    CLOSE = ValueConstant('')

class SERVICE(Constants):
    code_start = 101800

    EDIT = ValueConstant('')
    DELETE = ValueConstant('')
    PUBLISH = ValueConstant('')

class STATS(Constants):
    code_start = 102000

    DELETE = ValueConstant('')
    DELETE_DAY = ValueConstant('')

class HOT_DEPLOY(Constants):
    code_start = 102200
    CREATE = ValueConstant('')

    # New in 2.0
    AFTER_DEPLOY = ValueConstant('')

class SINGLETON(Constants):
    code_start = 102400
    CLOSE = ValueConstant('')

class MSG_NS(Constants):
    code_start = 102600

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')

class MSG_XPATH(Constants):
    code_start = 102800

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')

class MSG_JSON_POINTER(Constants):
    code_start = 103000

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')

class PUB_SUB_TOPIC(Constants):
    code_start = 103200

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')
    ADD_DEFAULT_PRODUCER = ValueConstant('')
    DELETE_DEFAULT_PRODUCER = ValueConstant('')

class PUB_SUB_PRODUCER(Constants):
    code_start = 103400

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')

class PUB_SUB_CONSUMER(Constants):
    code_start = 103600

    # New in 2.0
    CREATE = ValueConstant('')
    EDIT = ValueConstant('')
    DELETE = ValueConstant('')

class CLOUD(Constants):
    code_start = 103800

    # New in 2.0
    OPENSTACK_SWIFT_CREATE_EDIT = ValueConstant('')
    OPENSTACK_SWIFT_DELETE = ValueConstant('')

    # New in 2.0
    AWS_S3_CREATE_EDIT = ValueConstant('')
    AWS_S3_DELETE = ValueConstant('')

class NOTIF(Constants):
    code_start = 104000

    # New in 2.0
    RUN_NOTIFIER = ValueConstant('')

    CLOUD_OPENSTACK_SWIFT_CREATE_EDIT = ValueConstant('')
    CLOUD_OPENSTACK_SWIFT_DELETE = ValueConstant('')

    SQL_CREATE = ValueConstant('')
    SQL_EDIT = ValueConstant('')
    SQL_DELETE = ValueConstant('')

class SEARCH(Constants):
    code_start = 104200

    # New in 2.0
    ES_CREATE = ValueConstant('')
    ES_EDIT = ValueConstant('')
    ES_DELETE = ValueConstant('')
    ES_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    SOLR_CREATE = ValueConstant('')
    SOLR_EDIT = ValueConstant('')
    SOLR_DELETE = ValueConstant('')
    SOLR_CHANGE_PASSWORD = ValueConstant('')

class QUERY(Constants):
    code_start = 104400

    CASSANDRA_CREATE = ValueConstant('')
    CASSANDRA_EDIT = ValueConstant('')
    CASSANDRA_DELETE = ValueConstant('')
    CASSANDRA_CHANGE_PASSWORD = ValueConstant('')

class EMAIL(Constants):
    code_start = 104800

    # New in 2.0
    SMTP_CREATE = ValueConstant('')
    SMTP_EDIT = ValueConstant('')
    SMTP_DELETE = ValueConstant('')
    SMTP_CHANGE_PASSWORD = ValueConstant('')

    # New in 2.0
    IMAP_CREATE = ValueConstant('')
    IMAP_EDIT = ValueConstant('')
    IMAP_DELETE = ValueConstant('')
    IMAP_CHANGE_PASSWORD = ValueConstant('')

class RBAC(Constants):
    code_start = 105200

    # New in 2.0
    ROLE_CREATE = ValueConstant('')
    ROLE_EDIT = ValueConstant('')
    ROLE_DELETE = ValueConstant('')

    # New in 2.0
    CLIENT_ROLE_CREATE = ValueConstant('')
    CLIENT_ROLE_DELETE = ValueConstant('')

    # New in 2.0
    PERMISSION_CREATE = ValueConstant('')
    PERMISSION_EDIT = ValueConstant('')
    PERMISSION_DELETE = ValueConstant('')

    # New in 2.0
    ROLE_PERMISSION_CREATE = ValueConstant('')
    ROLE_PERMISSION_EDIT = ValueConstant('')
    ROLE_PERMISSION_DELETE = ValueConstant('')

code_to_name = {}

# To prevent 'RuntimeError: dictionary changed size during iteration'
item_name, item = None, None

for item_name, item in globals().items():
    if isclass(item) and issubclass(item, Constants) and item is not Constants:
        for idx, (attr, const) in enumerate(item.items()):
            const.value = str(item.code_start + idx)
            code_to_name[const.value.encode('utf-8')] = '{}_{}'.format(item_name, attr)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2015 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
import logging

# globre
from globre import match as globre_match

# Paste
from paste.util.converters import asbool

# Zato
from zato.common import FALSE_TRUE, TRUE_FALSE

logger = logging.getLogger(__name__)

class Matcher(object):
    def __init__(self):
        self.config = None
        self.items = {True:[], False:[]}
        self.order1 = None
        self.order2 = None
        self.is_allowed_cache = {}
        self.special_case = None

    def read_config(self, config):
        self.config = config
        order = config.get('order', FALSE_TRUE)
        self.order1, self.order2 = (True, False) if order == TRUE_FALSE else (False, True)

        for key, value in config.items():

            # Ignore meta key(s)
            if key == 'order':
                continue

            value = asbool(value)

            # Add new items
            self.items[value].append(key)

        # Now sort everything lexicographically, the way it will be used in run-time
        for key in self.items:
            self.items[key] = list(reversed(sorted(self.items[key])))

        for empty, non_empty in ((True, False), (False, True)):
            if not self.items[empty] and '*' in self.items[non_empty]:
                self.special_case = non_empty
                break

    def is_allowed(self, value):
        logger.debug('Cache:`%s`, value:`%s`', self.is_allowed_cache, value)

        if self.special_case is not None:
            return self.special_case

        try:
            return self.is_allowed_cache[value]
        except KeyError:
            _match = globre_match
            is_allowed = None

            for order in self.order1, self.order2:
                for pattern in self.items[order]:
                    if _match(pattern, value):
                        is_allowed = order

            # No match at all - we don't allow it in that case
            is_allowed = is_allowed if (is_allowed is not None) else False

            self.is_allowed_cache[value] = is_allowed
            return is_allowed



# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
from logging import getLogger

# gevent
from gevent.lock import RLock

logger = getLogger(__name__)

# ################################################################################################################################

UPDATES = 'CREATE', 'EDIT', 'DELETE', 'CHANGE_PASSWORD'

class Dispatcher(object):

    def __init__(self):
        self.lock = RLock()
        self.listeners = {}

    def _listen(self, event, callback, **opaque):
        self.listeners.setdefault(event, []).append((callback, opaque))

    def listen(self, *args, **kwargs):
        with self.lock:
            self._listen(*args, **kwargs)

    def listen_for_updates(self, msg, callback, **opaque):
        with self.lock:
            for name, value in msg.iteritems():
                for update in UPDATES:
                    if update in name:
                        self._listen(value.value, callback, **opaque)

    def notify(self, event, ctx):
        with self.lock:
            for ev, values in self.listeners.items():
                if ev == event:
                    for callback, opaque in values:
                        callback(event, ctx, **opaque)

# A singleton used throughout the whole application.
dispatcher = Dispatcher()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

# stdlib
from logging import getLogger

# Zato
from zato.common import engine_def, engine_def_sqlite, ZATO_NOT_GIVEN

logger = getLogger(__name__)

django_sa_mappings = {
    'NAME': 'db_name',
    'HOST': 'host',
    'PORT': 'port',
    'USER': 'username',
    'PASSWORD': 'password',
    'odb_type': 'engine',
    'db_type': 'engine',
}

cli_sa_mappings = {
    'odb_db_name': 'db_name',
    'odb_host': 'host',
    'odb_port': 'port',
    'odb_user': 'username',
    'odb_password': 'password',
    'odb_type': 'engine',
}

def get_engine_url(args):
    attrs = {}
    is_sqlite = False
    is_django = 'NAME' in args
    has_get = getattr(args, 'get', False)

    odb_type = getattr(args, 'odb_type', None)
    if odb_type:
        is_sqlite = odb_type == 'sqlite'
    else:
        is_sqlite = args.get('engine') == 'sqlite' or args.get('db_type') == 'sqlite'

    names = ('engine', 'username', 'password', 'host', 'port', 'name', 'db_name', 'db_type', 'sqlite_path', 'odb_type',
             'odb_user', 'odb_password', 'odb_host', 'odb_port', 'odb_db_name', 'odb_type', 'ENGINE', 'NAME', 'HOST', 'USER',
             'PASSWORD', 'PORT')

    for name in names:
        if has_get:
            attrs[name] = args.get(name, '')
        else:
            attrs[name] = getattr(args, name, '')

    # Re-map Django params into SQLAlchemy params
    if is_django:
        for name in django_sa_mappings:
            value = attrs.get(name, ZATO_NOT_GIVEN)
            if value != ZATO_NOT_GIVEN:
                if not value and (name in 'db_type', 'odb_type'):
                    continue
                attrs[django_sa_mappings[name]] = value

    # Zato CLI to SQLAlchemy
    if not attrs.get('engine'):
        for name in cli_sa_mappings:
            value = attrs.get(name, ZATO_NOT_GIVEN)
            if value != ZATO_NOT_GIVEN:
                attrs[cli_sa_mappings[name]] = value

    # Re-map server ODB params into SQLAlchemy params
    if attrs['engine'] == 'sqlite':
        db_name = attrs.get('db_name')
        sqlite_path = attrs.get('sqlite_path')

        if db_name:
            attrs['sqlite_path'] = db_name

        if sqlite_path:
            attrs['db_name'] = sqlite_path

    return (engine_def_sqlite if is_sqlite else engine_def).format(**attrs)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
import socket, ssl
from httplib import OK
from tempfile import NamedTemporaryFile
from threading import Thread
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer

# psutil
import psutil

# Zato
from zato.common import ZATO_OK
from zato.common.test.tls_material import ca_cert, server1_cert, server1_key

def get_free_port(start=20001, end=50000):
    taken = []
    for c in psutil.net_connections(kind='inet'):
        if c.status == psutil.CONN_LISTEN:
            taken.append(c.laddr[1])

    for port in xrange(start, end):
        if port not in taken:
            return port

class _HTTPHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.send_response(OK)
        self.send_header('Content-Type', 'application/json')
        self.wfile.write('\n')
        self.wfile.write('"{}"'.format(ZATO_OK))

    do_DELETE = do_OPTIONS = do_POST = do_PUT = do_PATCH = do_GET

class _TLSServer(HTTPServer):
    def __init__(self, cert_reqs, ca_cert):
        self.port = get_free_port()
        self.cert_reqs = cert_reqs
        self.ca_cert=None
        HTTPServer.__init__(self, ('0.0.0.0', self.port), _HTTPHandler)

    def server_bind(self):

        with NamedTemporaryFile(prefix='zato-tls', delete=False) as server1_key_tf:
            server1_key_tf.write(server1_key)
            server1_key_tf.flush()

            with NamedTemporaryFile(prefix='zato-tls', delete=False) as server1_cert_tf:
                server1_cert_tf.write(server1_cert)
                server1_cert_tf.flush()

                with NamedTemporaryFile(prefix='zato-tls', delete=False) as ca_cert_tf:
                    ca_cert_tf.write(ca_cert)
                    ca_cert_tf.flush()

                    self.socket = ssl.wrap_socket(
                        self.socket, server1_key_tf.name, server1_cert_tf.name, True, self.cert_reqs, ca_certs=ca_cert_tf.name)

                    if self.allow_reuse_address:
                        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

                    self.socket.bind(self.server_address)
                    self.server_address = self.socket.getsockname()

class TLSServer(Thread):
    def __init__(self, cert_reqs=ssl.CERT_NONE, ca_cert=None):
        Thread.__init__(self)
        self.setDaemon(True)

        self.server = None
        self.cert_reqs = cert_reqs
        self.ca_cert=None

    def get_port(self):
        return self.server.port

    def stop(self):
        self.server.server_close()

    def run(self):
        self.server = _TLSServer(self.cert_reqs, self.ca_cert)
        self.server.serve_forever()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2010 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

# stdlib
import os
from collections import OrderedDict
from copy import deepcopy
from cStringIO import StringIO
from httplib import responses
from string import Template
from sys import maxint
from traceback import format_exc

# boto
from boto.s3.key import Key

# Bunch
from bunch import Bunch

# candv
from candv import Constants, ValueConstant

# lxml
from lxml import etree
from lxml.objectify import ObjectPath as _ObjectPath


# ##############################################################################
# Version
# ##############################################################################

try:
    curdir = os.path.dirname(os.path.abspath(__file__))
    _version_py = os.path.normpath(os.path.join(curdir, '..', '..', '..', '..', '.version.py'))
    _locals = {}
    execfile(_version_py, _locals)
    version = _locals['version']
except IOError:
    version = '2.0.7'

# The namespace for use in all Zato's own services.
zato_namespace = 'https://zato.io/ns/20130518'
zato_ns_map = {None: zato_namespace}

# SQL ODB
engine_def = '{engine}://{username}:{password}@{host}:{port}/{db_name}'
engine_def_sqlite = 'sqlite:///{sqlite_path}'

# Convenience access functions and constants.

soapenv11_namespace = 'http://schemas.xmlsoap.org/soap/envelope/'
soapenv12_namespace = 'http://www.w3.org/2003/05/soap-envelope'

wsse_namespace = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd'
wsu_namespace = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd'

common_namespaces = {
    'soapenv':soapenv11_namespace,
    'wsse':wsse_namespace,
    'wsu':wsu_namespace,
    'zato':zato_namespace
}

soap_doc = Template("""<soap:Envelope xmlns:soap='%s'><soap:Body>$body</soap:Body></soap:Envelope>""" % soapenv11_namespace)

soap_body_path = '/soapenv:Envelope/soapenv:Body'
soap_body_xpath = etree.XPath(soap_body_path, namespaces=common_namespaces)

soap_fault_path = '/soapenv:Envelope/soapenv:Body/soapenv:Fault'
soap_fault_xpath = etree.XPath(soap_fault_path, namespaces=common_namespaces)

wsse_password_type_text = 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText'
supported_wsse_password_types = (wsse_password_type_text,)

wsse_username_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsse:UsernameToken/wsse:Username'
wsse_username_xpath = etree.XPath(wsse_username_path, namespaces=common_namespaces)

wsse_password_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsse:UsernameToken/wsse:Password'
wsse_password_xpath = etree.XPath(wsse_password_path, namespaces=common_namespaces)

wsse_password_type_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsse:UsernameToken/wsse:Password/@Type'
wsse_password_type_xpath = etree.XPath(wsse_password_type_path, namespaces=common_namespaces)

wsse_nonce_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsse:UsernameToken/wsse:Nonce'
wsse_nonce_xpath = etree.XPath(wsse_nonce_path, namespaces=common_namespaces)

wsu_username_created_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsse:UsernameToken/wsu:Created'
wsu_username_created_xpath = etree.XPath(wsu_username_created_path, namespaces=common_namespaces)

wsu_expires_path = '/soapenv:Envelope/soapenv:Header/wsse:Security/wsu:Timestamp/wsu:Expires'
wsu_expires_xpath = etree.XPath(wsu_expires_path, namespaces=common_namespaces)

wsse_username_objectify = '{}Security'.format(wsse_namespace)
wsse_username_token_objectify = '{}UsernameToken'.format(wsse_namespace)

zato_data_path = soap_data_path = '/soapenv:Envelope/soapenv:Body/*[1]'
zato_data_xpath = soap_data_xpath = etree.XPath(zato_data_path, namespaces=common_namespaces)

zato_result_path = '//zato:zato_env/zato:result'
zato_result_xpath = etree.XPath(zato_result_path, namespaces=common_namespaces)

zato_cid_path = '//zato:zato_env/zato:cid'
zato_cid_xpath = etree.XPath(zato_result_path, namespaces=common_namespaces)

zato_details_path = '//zato:zato_env/zato:details'
zato_details_xpath = etree.XPath(zato_details_path, namespaces=common_namespaces)

SECRET_SHADOW = '******'

# TRACE1 logging level, even more details than DEBUG
TRACE1 = 6

SECONDS_IN_DAY = 86400 # 60 seconds * 60 minutes * 24 hours (and we ignore leap seconds)

#iso8601_format = '%Y-%m-%dT%H:%M:%S.%fZ'
scheduler_date_time_format = '%Y-%m-%d %H:%M:%S'
soap_date_time_format = '%Y-%m-%dT%H:%M:%S.%fZ'

ACCESS_LOG_DT_FORMAT = '%d/%b/%Y:%H:%M:%S %z'

# TODO: Classes that have this attribute defined (no matter the value) will not be deployed
# onto servers.
DONT_DEPLOY_ATTR_NAME = 'zato_dont_import'

# A convenient constant used in several places, simplifies passing around
# arguments which are, well, not given (as opposed to being None, an empty string etc.)
ZATO_NOT_GIVEN = b'ZATO_NOT_GIVEN'

# Also used in a couple of places.
ZATO_OK = 'ZATO_OK'
ZATO_ERROR = 'ZATO_ERROR'
ZATO_WARNING = 'ZATO_WARNING'
ZATO_NONE = b'ZATO_NONE'

# Default HTTP method outgoing connections use to ping resources
# TODO: Move it to MISC
DEFAULT_HTTP_PING_METHOD = 'HEAD'

# Default size of an outgoing HTTP connection's pool (plain, SOAP, any).
# This is a per-outconn setting
# TODO: Move it to MISC
DEFAULT_HTTP_POOL_SIZE = 20

# Used when there's a need for encrypting/decrypting a well-known data.
# TODO: Move it to MISC
ZATO_CRYPTO_WELL_KNOWN_DATA = 'ZATO'

# https://tools.ietf.org/html/rfc6585
TOO_MANY_REQUESTS = 429

HTTP_RESPONSES = deepcopy(responses)
HTTP_RESPONSES[TOO_MANY_REQUESTS] = 'Too Many Requests'

# Pattern matching order
TRUE_FALSE = 'true_false'
FALSE_TRUE = 'false_true'

# Queries to use in pinging the databases.
ping_queries = {
    'access': 'SELECT 1',
    'db2': 'SELECT current_date FROM sysibm.sysdummy1',
    'firebird': 'SELECT current_timestamp FROM rdb$database',
    'informix': 'SELECT 1 FROM systables WHERE tabid=1',
    'mssql': 'SELECT 1',
    'mysql+pymysql': 'SELECT 1+1',
    'oracle': 'SELECT 1 FROM dual',
    'postgresql': 'SELECT 1',
    'postgresql+pg8000': 'SELECT 1',
    'sqlite': 'SELECT 1',
}

# All URL types Zato understands.
class URL_TYPE(object):
    SOAP = 'soap'
    PLAIN_HTTP = 'plain_http'

    class __metaclass__(type):
        def __iter__(self):
            return iter((self.SOAP, self.PLAIN_HTTP))

# Whether WS-Security passwords are transmitted in clear-text or not.
ZATO_WSS_PASSWORD_CLEAR_TEXT = Bunch(name='clear_text', label='Clear text')
ZATO_WSS_PASSWORD_TYPES = {
    ZATO_WSS_PASSWORD_CLEAR_TEXT.name:ZATO_WSS_PASSWORD_CLEAR_TEXT.label,
}

ZATO_FIELD_OPERATORS = {
    'is-equal-to': '==',
    'is-not-equal-to': '!=',
}

ZMQ_OUTGOING_TYPES = ('PUSH', 'PUB')
ZMQ_CHANNEL_TYPES = ('PULL', 'SUB')

ZATO_ODB_POOL_NAME = 'ZATO_ODB'

SOAP_VERSIONS = ('1.1', '1.2')
SOAP_CHANNEL_VERSIONS = ('1.1',)

    
class SEARCH:
    class ES:
        class DEFAULTS(Constants):
            BODY_AS = ValueConstant('POST')
            HOSTS = ValueConstant('127.0.0.1:9200\n')

    class SOLR:
        class DEFAULTS(Constants):
            ADDRESS = ValueConstant('http://127.0.0.1:8983/solr')
            PING_PATH = ValueConstant('/solr/admin/ping')
            TIMEOUT = ValueConstant('10')
            POOL_SIZE = ValueConstant('5')

class SEC_DEF_TYPE:
    APIKEY = 'apikey'
    AWS = 'aws'
    BASIC_AUTH = 'basic_auth'
    NTLM = 'ntlm'
    OAUTH = 'oauth'
    OPENSTACK = 'openstack'
    TECH_ACCOUNT = 'tech_acc'
    TLS_CHANNEL_SEC = 'tls_channel_sec'
    TLS_KEY_CERT = 'tls_key_cert'
    WSS = 'wss'
    XPATH_SEC = 'xpath_sec'

SEC_DEF_TYPE_NAME = {
    SEC_DEF_TYPE.APIKEY: 'API key',
    SEC_DEF_TYPE.AWS: 'AWS',
    SEC_DEF_TYPE.BASIC_AUTH: 'HTTP Basic Auth',
    SEC_DEF_TYPE.NTLM: 'NTLM',
    SEC_DEF_TYPE.OAUTH: 'OAuth 1.0',
    SEC_DEF_TYPE.OPENSTACK: 'OpenStack',
    SEC_DEF_TYPE.TECH_ACCOUNT: 'Tech account',
    SEC_DEF_TYPE.WSS: 'WS-Security',
    SEC_DEF_TYPE.TLS_CHANNEL_SEC: 'TLS channel',
    SEC_DEF_TYPE.TLS_KEY_CERT: 'TLS key/cert',
    SEC_DEF_TYPE.XPATH_SEC: 'XPath',
}

# Name of the scheduler's job that will ensure a singleton server is always
# available in a cluster.
ENSURE_SINGLETON_JOB = 'zato.server.ensure-cluster-wide-singleton'

DEFAULT_STATS_SETTINGS = {
    'scheduler_per_minute_aggr_interval':60,
    'scheduler_raw_times_interval':90,
    'scheduler_raw_times_batch':99999,
    'atttention_slow_threshold':2000,
    'atttention_top_threshold':10,
}

class BATCH_DEFAULTS:
    PAGE_NO = 1
    SIZE = 25
    MAX_SIZE = 1000

class NameId(object):
    """ Wraps both an attribute's name and its ID.
    """
    def __init__(self, name, id):
        self.name = name
        self.id = id

class NotGiven(object):
    pass # A marker for lazily-initialized attributes

class Attrs(type):
    """ A container for class attributes that can be queried for an existence
    of an attribute using the .has class-method.
    """
    attrs = NotGiven

    @classmethod
    def has(cls, attr):
        if cls.attrs is NotGiven:
            cls.attrs = []
            for cls_attr in dir(cls):
                if cls_attr == cls_attr.upper():
                    cls.attrs.append(getattr(cls, cls_attr))

        return attr in cls.attrs

class DATA_FORMAT(Attrs):
    DICT = 'dict'
    XML = 'xml'
    JSON = 'json'
    CSV = 'csv'
    POST = 'post'

    class __metaclass__(type):
        def __iter__(self):
            # Note that DICT isn't included because it's never exposed to external world as-is,
            # it's only used so that services can invoke each other directly
            return iter((self.XML, self.JSON, self.CSV, self.POST))

# TODO: SIMPLE_IO.FORMAT should be done away with in favour of plain DATA_FORMAT
class SIMPLE_IO:
    class FORMAT(Attrs):
        XML = DATA_FORMAT.XML
        JSON = DATA_FORMAT.JSON

    class INT_PARAMETERS:
        VALUES = ['id']
        SUFFIXES = ['_id', '_count', '_size', '_timeout']

    class BOOL_PARAMETERS:
        SUFFIXES = ['is_', 'needs_', 'should_']

class DEPLOYMENT_STATUS(Attrs):
    DEPLOYED = 'deployed'
    AWAITING_DEPLOYMENT = 'awaiting-deployment'
    IGNORED = 'ignored'

class SERVER_JOIN_STATUS(Attrs):
    ACCEPTED = 'accepted'

class SERVER_UP_STATUS(Attrs):
    RUNNING = 'running'
    CLEAN_DOWN = 'clean-down'

class KVDB(Attrs):
    SEPARATOR = ':::'

    DICTIONARY_ITEM = 'zato:kvdb:data-dict:item'
    DICTIONARY_ITEM_ID = DICTIONARY_ITEM + ':id' # ID of the last created dictionary ID, always increasing.

    LOCK_PREFIX = 'zato:lock:'

    LOCK_SERVER_PREFIX = '{}server:'.format(LOCK_PREFIX)
    LOCK_SERVER_ALREADY_DEPLOYED = '{}already-deployed:'.format(LOCK_SERVER_PREFIX)
    LOCK_SERVER_STARTING = '{}starting:'.format(LOCK_SERVER_PREFIX)

    LOCK_PACKAGE_PREFIX = '{}package:'.format(LOCK_PREFIX)
    LOCK_PACKAGE_UPLOADING = '{}uploading:'.format(LOCK_PACKAGE_PREFIX)
    LOCK_PACKAGE_ALREADY_UPLOADED = '{}already-uploaded:'.format(LOCK_PACKAGE_PREFIX)

    LOCK_DELIVERY = '{}delivery:'.format(LOCK_PREFIX)
    LOCK_DELIVERY_AUTO_RESUBMIT = '{}auto-resubmit:'.format(LOCK_DELIVERY)

    LOCK_SERVICE_PREFIX = '{}service:'.format(LOCK_PREFIX)

    LOCK_FANOUT_PATTERN = '{}fanout:{{}}'.format(LOCK_PREFIX)
    LOCK_PARALLEL_EXEC_PATTERN = '{}parallel-exec:{{}}'.format(LOCK_PREFIX)

    LOCK_ASYNC_INVOKE_WITH_TARGET_PATTERN = '{}async-invoke-with-pattern:{{}}:{{}}'.format(LOCK_PREFIX)

    TRANSLATION = 'zato:kvdb:data-dict:translation'
    TRANSLATION_ID = TRANSLATION + ':id'

    SERVICE_USAGE = 'zato:stats:service:usage:'
    SERVICE_TIME_BASIC = 'zato:stats:service:time:basic:'
    SERVICE_TIME_RAW = 'zato:stats:service:time:raw:'
    SERVICE_TIME_RAW_BY_MINUTE = 'zato:stats:service:time:raw-by-minute:'
    SERVICE_TIME_AGGREGATED_BY_MINUTE = 'zato:stats:service:time:aggr-by-minute:'
    SERVICE_TIME_AGGREGATED_BY_HOUR = 'zato:stats:service:time:aggr-by-hour:'
    SERVICE_TIME_AGGREGATED_BY_DAY = 'zato:stats:service:time:aggr-by-day:'
    SERVICE_TIME_AGGREGATED_BY_MONTH = 'zato:stats:service:time:aggr-by-month:'
    SERVICE_TIME_SLOW = 'zato:stats:service:time:slow:'

    SERVICE_SUMMARY_PREFIX_PATTERN = 'zato:stats:service:summary:{}:'
    SERVICE_SUMMARY_BY_DAY = 'zato:stats:service:summary:by-day:'
    SERVICE_SUMMARY_BY_WEEK = 'zato:stats:service:summary:by-week:'
    SERVICE_SUMMARY_BY_MONTH = 'zato:stats:service:summary:by-month:'
    SERVICE_SUMMARY_BY_YEAR = 'zato:stats:service:summary:by-year:'

    REQ_RESP_SAMPLE = 'zato:req-resp:sample:'
    RESP_SLOW = 'zato:resp:slow:'

    DELIVERY_PREFIX = 'zato:delivery:'
    DELIVERY_BY_TARGET_PREFIX = '{}by-target:'.format(DELIVERY_PREFIX)

    FANOUT_COUNTER_PATTERN = 'zato:fanout:counter:{}'
    FANOUT_DATA_PATTERN = 'zato:fanout:data:{}'

    PARALLEL_EXEC_COUNTER_PATTERN = 'zato:parallel-exec:counter:{}'
    PARALLEL_EXEC_DATA_PATTERN = 'zato:parallel-exec:data:{}'

    ASYNC_INVOKE_PROCESSED_FLAG_PATTERN = 'zato:async-invoke-with-pattern:processed:{}:{}'
    ASYNC_INVOKE_PROCESSED_FLAG = '1'

class SCHEDULER:

    class JOB_TYPE(Attrs):
        ONE_TIME = 'one_time'
        INTERVAL_BASED = 'interval_based'
        CRON_STYLE = 'cron_style'

    class ON_MAX_RUNS_REACHED:
        DELETE = 'delete'
        INACTIVATE = 'inactivate'

class CHANNEL(Attrs):
    AMQP = 'amqp'
    AUDIT = 'audit'                                       # New in 2.0
    DELIVERY = 'delivery'                                 # New in 2.0
    FANOUT_CALL = 'fanout-call'                           # New in 2.0
    FANOUT_ON_FINAL = 'fanout-on-final'                   # New in 2.0
    FANOUT_ON_TARGET = 'fanout-on-target'                 # New in 2.0
    HTTP_SOAP = 'http-soap'
    INTERNAL_CHECK = 'internal-check'                     # New in 2.0
    INVOKE = 'invoke'
    INVOKE_ASYNC = 'invoke-async'
    INVOKE_ASYNC_CALLBACK = 'invoke-async-callback'       # New in 2.0
    JMS_WMQ = 'jms-wmq'
    NOTIFIER_RUN = 'notifier-run'                         # New in 2.0
    NOTIFIER_TARGET = 'notifier-target'                   # New in 2.0
    PARALLEL_EXEC_CALL = 'parallel-exec-call'             # New in 2.0
    PARALLEL_EXEC_ON_TARGET = 'parallel-exec-on-target'   # New in 2.0
    PUBLISH = 'publish'                                   # New in 2.0
    SCHEDULER = 'scheduler'
    SCHEDULER_AFTER_ONE_TIME = 'scheduler-after-one-time' # New in 2.0
    STARTUP_SERVICE = 'startup-service'                   # New in 2.0
    WORKER = 'worker'                                     # New in 2.0
    ZMQ = 'zmq'

class INVOCATION_TARGET(Attrs):
    CHANNEL_AMQP = 'channel-amqp'
    CHANNEL_WMQ = 'channel-wmq'
    CHANNEL_ZMQ = 'channel-zmq'
    OUTCONN_AMQP = 'outconn-amqp'
    OUTCONN_WMQ = 'outconn-wmq'
    OUTCONN_ZMQ = 'outconn-zmq'
    SERVICE = 'service'

class DELIVERY_HISTORY_ENTRY(Attrs):
    ENTERED_IN_DOUBT = b'entered-in-doubt'
    ENTERED_IN_PROGRESS = b'entered-in-progress'
    ENTERED_CONFIRMED = b'entered-confirmed'
    ENTERED_FAILED = b'entered-failed'
    ENTERED_RETRY = b'entered-retry'
    NONE = b'(None)'
    SENT_FROM_SOURCE = b'sent-from-source'
    SENT_FROM_SOURCE_RESUBMIT = b'sent-from-source-resubmit'
    SENT_FROM_SOURCE_RESUBMIT_AUTO = b'sent-from-source-resubmit-auto'
    TARGET_OK = b'target-ok'
    TARGET_FAILURE = b'target-failure'
    UPDATED = b'updated'

class DELIVERY_STATE(Attrs):
    IN_DOUBT = 'in-doubt'
    IN_PROGRESS_ANY = 'in-progress-any' # A wrapper for all in-progress-* states
    IN_PROGRESS_RESUBMITTED = 'in-progress-resubmitted'
    IN_PROGRESS_RESUBMITTED_AUTO = 'in-progress-resubmitted-auto'
    IN_PROGRESS_STARTED = 'in-progress'
    IN_PROGRESS_TARGET_OK = 'in-progress-target-ok'
    IN_PROGRESS_TARGET_FAILURE = 'in-progress-target-failure'
    CONFIRMED = 'confirmed'
    FAILED = 'failed'
    UNKNOWN = 'unknown'

class DELIVERY_COUNTERS(Attrs):
    IN_DOUBT = 'in_doubt_count'
    IN_PROGRESS = 'in_progress_count'
    CONFIRMED = 'confirmed_count'
    FAILED = 'failed_count'
    TOTAL = 'total_count'

class DELIVERY_CALLBACK_INVOKER(Attrs):
    SOURCE = 'source'
    TARGET = 'target'

class BROKER:
    DEFAULT_EXPIRATION = 15 # In seconds

class MISC:
    DEFAULT_HTTP_TIMEOUT=10
    DEFAULT_AUDIT_BACK_LOG = 24 * 60 # 24 hours * 60 days ≅ 2 months
    DEFAULT_AUDIT_MAX_PAYLOAD = 0 # Using 0 means there's no limit
    OAUTH_SIG_METHODS = ['HMAC-SHA1', 'PLAINTEXT']
    PIDFILE = 'pidfile'
    SEPARATOR = ':::'

class ADAPTER_PARAMS:
    APPLY_AFTER_REQUEST = 'apply-after-request'
    APPLY_BEFORE_REQUEST = 'apply-before-request'

class AUDIT_LOG:
    REPLACE_WITH = SECRET_SHADOW

class INFO_FORMAT:
    DICT = 'dict'
    TEXT = 'text'
    JSON = 'json'
    YAML = 'yaml'

class MSG_MAPPER:
    DICT_TO_DICT = 'dict-to-dict'
    DICT_TO_XML = 'dict-to-xml'
    XML_TO_DICT = 'xml-to-dict'
    XML_TO_XML = 'xml-to-xml'

class CLOUD:
    class OPENSTACK:
        class SWIFT:
            class DEFAULTS:
                AUTH_VERSION = '1'
                BACKOFF_STARTING = 1
                BACKOFF_MAX = 64
                POOL_SIZE = 5
                RETRIES = 5

    class AWS:
        class S3:
            class STORAGE_CLASS:
                STANDARD = 'STANDARD'
                REDUCED_REDUNDANCY = 'REDUCED_REDUNDANCY'
                GLACIER = 'GLACIER'
                DEFAULT = STANDARD

                class __metaclass__(type):
                    def __iter__(self):
                        return iter((self.STANDARD, self.REDUCED_REDUNDANCY, self.GLACIER))

            class DEFAULTS:
                ADDRESS = 'https://s3.amazonaws.com/'
                CONTENT_TYPE = Key.DefaultContentType
                DEBUG_LEVEL = 0
                POOL_SIZE = 5
                PROVIDER = 'aws'

class URL_PARAMS_PRIORITY:
    PATH_OVER_QS = 'path-over-qs'
    QS_OVER_PATH = 'qs-over-path'
    DEFAULT = QS_OVER_PATH

    class __metaclass__(type):
        def __iter__(self):
            return iter((self.PATH_OVER_QS, self.QS_OVER_PATH, self.DEFAULT))

class PARAMS_PRIORITY:
    CHANNEL_PARAMS_OVER_MSG = 'channel-params-over-msg'
    MSG_OVER_CHANNEL_PARAMS = 'msg-over-channel-params'
    DEFAULT = CHANNEL_PARAMS_OVER_MSG

    class __metaclass__(type):
        def __iter__(self):
            return iter((self.CHANNEL_PARAMS_OVER_MSG, self.MSG_OVER_CHANNEL_PARAMS, self.DEFAULT))

class NONCE_STORE:
    KEY_PATTERN = 'zato:nonce-store:{}:{}' # E.g. zato:nonce-store:oauth:27
    DEFAULT_MAX_LOG = 25000

class MSG_PATTERN_TYPE:
    JSON_POINTER = NameId('JSONPointer', 'json-pointer')
    XPATH = NameId('XPath', 'xpath')
    
    class __metaclass__(type):
        def __iter__(self):
            return iter((self.JSON_POINTER, self.XPATH))

class HTTP_SOAP_SERIALIZATION_TYPE:
    STRING_VALUE = NameId('String', 'string')
    SUDS = NameId('Suds', 'suds')
    DEFAULT = STRING_VALUE

    class __metaclass__(type):
        def __iter__(self):
            return iter((self.STRING_VALUE, self.SUDS))

class PUB_SUB:
    PRIORITY_MIN = 1
    PRIORITY_MAX = 9

    DEFAULT_PRIORITY = 5
    DEFAULT_MIME_TYPE = 'text/plain'
    DEFAULT_EXPIRATION = 60.0 # In seconds
    DEFAULT_GET_MAX_BATCH_SIZE = 100
    DEFAULT_IS_FIFO = True
    DEFAULT_MAX_DEPTH = 500
    DEFAULT_MAX_BACKLOG = 1000

    class QUEUE_TYPE:
        MESSAGE = 'message'
        IN_FLIGHT = 'in-flight'

    class GET_DIR:
        FIFO = 'fifo'
        LIFO = 'lifo'

    class MOVE_RESULT:
        MOVED = 'moved'
        OVERFLOW = 'overflow'


    class CALLBACK_TYPE:
        OUTCONN_PLAIN_HTTP = 'outconn-plain-http'
        OUTCONN_SOAP = 'outconn-soap'

    class DELIVERY_MODE:
        PULL = NameId('Pull', 'pull')
        CALLBACK_URL = NameId('Callback URL', 'callback-url')

        class __metaclass__(type):
            def __iter__(self):
                return iter((self.PULL, self.CALLBACK_URL))

    class MESSAGE_SOURCE:
        TOPIC = NameId('Topic', 'topic')
        CONSUMER_QUEUE = NameId('Consumer queue', 'consumer-queue')
    
        class __metaclass__(type):
            def __iter__(self):
                return iter((self.TOPIC, self.CONSUMER_QUEUE))

    class GET_FORMAT:
        OBJECT = NameId('Object', 'object')
        JSON = NameId('JSON', 'json')
        XML = NameId('XML', 'xml')
        DEFAULT = JSON
    
        class __metaclass__(type):
            def __iter__(self):
                return iter((self.OBJECT, self.JSON, self.XML))

    class URL_ITEM_TYPE:
        TOPIC = NameId('Topic', 'topic')
        MESSAGES = NameId('Messages', 'msg')
    
        class __metaclass__(type):
            def __iter__(self):
                return iter((self.TOPIC.id, self.MESSAGES.id))

class EMAIL:
    class DEFAULT:
        TIMEOUT = 10
        PING_ADDRESS = 'invalid@invalid'
        GET_CRITERIA = 'UNSEEN'
        IMAP_DEBUG_LEVEL = 0

    class IMAP:
        class MODE(Constants):
            PLAIN = ValueConstant('plain')
            SSL = ValueConstant('ssl')

    class SMTP:
        class MODE(Constants):
            PLAIN = ValueConstant('plain')
            SSL = ValueConstant('ssl')
            STARTTLS = ValueConstant('starttls')

class NOTIF:
    class DEFAULT:
        CHECK_INTERVAL = 5 # In seconds
        CHECK_INTERVAL_SQL = 600 # In seconds
        NAME_PATTERN = '**'
        GET_DATA_PATTERN = '**'

    class TYPE:
        OPENSTACK_SWIFT = 'openstack_swift'
        SQL = 'sql'

class CASSANDRA:
    class DEFAULT(Constants):
        CONTACT_POINTS = ValueConstant('127.0.0.1\n')
        EXEC_SIZE = ValueConstant(2)
        PORT = ValueConstant(9042)
        PROTOCOL_VERSION = ValueConstant(2)
        KEYSPACE = ValueConstant('not-set')

    class COMPRESSION(Constants):
        DISABLED = ValueConstant('disabled')
        ENABLED_NEGOTIATED = ValueConstant('enabled-negotiated')
        ENABLED_LZ4 = ValueConstant('enabled-lz4')
        ENABLED_SNAPPY = ValueConstant('enabled-snappy')

class TLS:
    # All the BEGIN/END blocks we don't want to store in logs.
    # Taken from https://github.com/openssl/openssl/blob/master/crypto/pem/pem.h
    # Note that the last one really is empty to denote 'BEGIN PRIVATE KEY' alone.
    BEGIN_END = ('ANY ', 'RSA ', 'DSA ', 'EC ', 'ENCRYPTED ', '')

    # Directories in a server's config/tls directory keeping the material
    DIR_CA_CERTS = 'ca-certs'
    DIR_KEYS_CERTS = 'keys-certs'

class ODOO:
    class DEFAULT:
        PORT = 8069
        POOL_SIZE = 3

    class PROTOCOL:
        XML_RPC = NameId('XML-RPC', 'xmlrpc')
        XML_RPCS = NameId('XML-RPCS', 'xmlrpcs')
        JSON_RPC = NameId('JSON-RPC', 'jsonrpc')
        JSON_RPCS = NameId('JSON-RPCS', 'jsonrpcs')

        class __metaclass__(type):
            def __iter__(self):
                return iter((self.XML_RPC, self.XML_RPCS, self.JSON_RPC, self.JSON_RPCS))

# Need to use such a constant because we can sometimes be interested in setting
# default values which evaluate to boolean False.
NO_DEFAULT_VALUE = 'NO_DEFAULT_VALUE'

ZATO_INFO_FILE = b'.zato-info'

class path(object):
    def __init__(self, path, raise_on_not_found=False, ns='', text_only=False):
        self.path = path
        self.ns = ns
        self.raise_on_not_found = raise_on_not_found
        self.text_only = text_only
        self.children_only = False
        self.children_only_idx = None

    def get_from(self, elem):
        if self.ns:
            _path = '{{{}}}{}'.format(self.ns, self.path)
        else:
            _path = self.path
        try:
            if self.children_only:
                elem = elem.getchildren()[self.children_only_idx]
            value = _ObjectPath(_path)(elem)
            if self.text_only:
                return value.text
            return value
        except(ValueError, AttributeError), e:
            if self.raise_on_not_found:
                raise ParsingException(None, format_exc(e))
            else:
                return None

class zato_path(path):
    def __init__(self, path, raise_on_not_found=False, text_only=False):
        super(zato_path, self).__init__(path, raise_on_not_found, zato_namespace, text_only)
        self.children_only = True
        self.children_only_idx = 1 # 0 is zato_env

class ZatoException(Exception):
    """ Base class for all Zato custom exceptions.
    """
    def __init__(self, cid=None, msg=None):
        super(ZatoException, self).__init__(msg)
        self.cid = cid
        self.msg = msg

class ClientSecurityException(ZatoException):
    """ An exception for signalling errors stemming from security problems
    on the client side, such as invalid username or password.
    """

class ConnectionException(ZatoException):
    """ Encountered a problem with an external connections, such as to AMQP brokers.
    """

class TimeoutException(ConnectionException):
    pass

class HTTPException(ZatoException):
    """ Raised when the underlying error condition can be easily expressed
    as one of the HTTP status codes.
    """
    def __init__(self, cid, msg, status):
        super(HTTPException, self).__init__(cid, msg)
        self.status = status
        self.reason = HTTP_RESPONSES[status]

    def __repr__(self):
        return '<{} at {} cid:`{}`, status:`{}`, msg:`{}`>'.format(
            self.__class__.__name__, hex(id(self)), self.cid, self.status, self.msg)

class ParsingException(ZatoException):
    """ Raised when the error is to do with parsing of documents, such as an input
    XML document.
    """

class NoDistributionFound(ZatoException):
    """ Raised when an attempt is made to import services from a Distutils2 archive
    or directory but they don't contain a proper Distutils2 distribution.
    """
    def __init__(self, path):
        super(NoDistributionFound, self).__init__(None, 'No Disutils distribution in path:[{}]'.format(path))

class Inactive(ZatoException):
    """ Raised when an attempt was made to use an inactive resource, such
    as an outgoing connection or a channel.
    """
    def __init__(self, name):
        super(Inactive, self).__init__(None, '`{}` is inactive'.format(name))

class SourceInfo(object):
    """ A bunch of attributes dealing the service's source code.
    """
    def __init__(self):
        self.source = None
        self.source_html = None
        self.path = None
        self.hash = None
        self.hash_method = None
        self.server_name = None

class StatsElem(object):
    """ A single element of a statistics query result concerning a particular service.
    All values make sense only within the time interval of the original query, e.g. a 'min_resp_time'
    may be 18 ms in this element because it represents statistics regarding, say,
    the last hour yet in a different period the 'min_resp_time' may be a completely
    different value. Likewise, 'all' in the description of parameters below means
    'all that matched given query criteria' rather than 'all that ever existed'.

    service_name - name of the service this element describes
    usage - how many times the service has been invoked
    mean - an arithmetical average of all the mean response times  (in ms)
    rate - usage rate in requests/s (up to 1 decimal point)
    time - time spent by this service on processing the messages (in ms)
    usage_trend - a CSV list of values representing the service usage
    usage_trend_int - a list of integers representing the service usage
    mean_trend - a CSV list of values representing mean response times (in ms)
    mean_trend_int - a list of integers representing mean response times (in ms)
    min_resp_time - minimum service response time (in ms)
    max_resp_time - maximum service response time (in ms)
    all_services_usage - how many times all the services have been invoked
    all_services_time - how much time all the services spent on processing the messages (in ms)
    mean_all_services - an arithmetical average of all the mean response times  of all services (in ms)
    usage_perc_all_services - this service's usage as a percentage of all_services_usage (up to 2 decimal points)
    time_perc_all_services - this service's share as a percentage of all_services_time (up to 2 decimal points)
    expected_time_elems - an OrderedDict of all the time slots mapped to a mean time and rate
    temp_rate - a temporary place for keeping request rates, needed to get a weighted mean of uneven execution periods
    temp_mean - just like temp_rate but for mean response times
    temp_mean_count - how many periods containing a mean rate there were
    """
    def __init__(self, service_name=None, mean=None):
        self.service_name = service_name
        self.usage = 0
        self.mean = mean
        self.rate = 0.0
        self.time = 0
        self.usage_trend_int = []
        self.mean_trend_int = []
        self.min_resp_time = maxint # Assuming that there sure will be at least one response time lower than that
        self.max_resp_time = 0
        self.all_services_usage = 0
        self.all_services_time = 0
        self.mean_all_services = 0
        self.usage_perc_all_services = 0
        self.time_perc_all_services = 0
        self.expected_time_elems = OrderedDict()
        self.temp_rate = 0
        self.temp_mean = 0
        self.temp_mean_count = 0

    def get_attrs(self, ignore=[]):
        for attr in dir(self):
            if attr.startswith('__') or attr.startswith('temp_') or callable(getattr(self, attr)) or attr in ignore:
                continue
            yield attr

    def to_dict(self, ignore=None):
        if not ignore:
            ignore = ['expected_time_elems', 'mean_trend_int', 'usage_trend_int']
        return {attr: getattr(self, attr) for attr in self.get_attrs(ignore)}

    @staticmethod
    def from_json(item):
        stats_elem = StatsElem()
        for k, v in item.items():
            setattr(stats_elem, k, v)

        return stats_elem

    @staticmethod
    def from_xml(item):
        stats_elem = StatsElem()
        for child in item.getchildren():
            setattr(stats_elem, child.xpath('local-name()'), child.pyval)

        return stats_elem

    def __repr__(self):
        buff = StringIO()
        buff.write('<{} at {} '.format(self.__class__.__name__, hex(id(self))))

        attrs = ('{}=[{}]'.format(attr, getattr(self, attr)) for attr in self.get_attrs())
        buff.write(', '.join(attrs))

        buff.write('>')

        value = buff.getvalue()
        buff.close()

        return value

    def __iadd__(self, other):
        self.max_resp_time = max(self.max_resp_time, other.max_resp_time)
        self.min_resp_time = min(self.min_resp_time, other.min_resp_time)
        self.usage += other.usage

        return self

    def __bool__(self):
        return bool(self.service_name) # Empty stats_elems won't have a service name set

# ################################################################################################################################

class SMTPMessage(object):
    def __init__(self, from_=None, to=None, subject='', body='', attachments=None, cc=None, bcc=None, is_html=False, headers=None,
            charset='utf8', is_rfc2231=True):
        self.from_ = from_
        self.to = to
        self.subject = subject
        self.body = body
        self.attachments = attachments or []
        self.cc = cc
        self.bcc = bcc
        self.is_html = is_html
        self.headers = headers or {}
        self.charset = charset
        self.is_rfc2231 = is_rfc2231

    def attach(self, name, contents):
        self.attachments.append({'name':name, 'contents':contents})

class IMAPMessage(object):
    def __init__(self, uid, conn, data):
        self.uid = uid
        self.conn = conn
        self.data = data

    def __repr__(self):
        return '<{} at {}, uid:`{}`, conn.config:`{}`>'.format(
            self.__class__.__name__, hex(id(self)), self.uid, self.conn.config_no_sensitive)

    def delete(self):
        self.conn.delete(self.uid)

    def mark_seen(self):
        self.conn.mark_seen(self.uid)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2014 Dariusz Suchojad <dsuch at zato.io>

Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from butler import Butler

class DictNav(Butler):
    """ Easy navigation over Python dicts.
    """
    def get(self, path, default=None):
        """ Overridden from the base class so it doesn't display warnings on non-existing paths.
        """
        try:
            return self[path]
        except (LookupError, TypeError):
            return default

    def has_key(self, key, nested=True):
        if nested:
            return bool(self.findall(key, True))

        return key in self.obj

    def has_path(self, path):
        return super(DictNav, self).path_exists(path)

# For now they are the same class but may part at some time
ListNav = DictNav